import logging
from collections import defaultdict
from typing import Tuple, Optional

from multiqc.plots.table_object import DataTable
from multiqc.utils import config, mqc_colour, util_functions, report

logger = logging.getLogger(__name__)


def plot(dt: DataTable) -> str:
    from multiqc.plots.plotly import violin

    return violin.plot(dt, show_table_by_default=True)


def make_table(dt: DataTable, violin_id: Optional[str] = None) -> Tuple[str, str]:
    """
    Build the HTML needed for a MultiQC table.
    :param dt: MultiQC datatable object
    :param violin_id: optional, will add a button to switch to a violin plot with this ID
    """

    t_headers = dict()
    t_modal_headers = dict()
    t_rows = dict()
    t_rows_empty = dict()
    dt.raw_vals = defaultdict(lambda: dict())
    empty_cells = dict()
    hidden_cols = 1
    table_title = dt.pconfig.get("table_title")
    if table_title is None:
        table_title = dt.id.replace("_", " ").title()

    for idx, k, header in dt.get_headers_in_order():
        rid = header["rid"]

        # Build the table header cell
        shared_key = ""
        if header.get("shared_key", None) is not None:
            shared_key = f" data-shared-key={header['shared_key']}"

        hide = ""
        muted = ""
        checked = ' checked="checked"'
        if header.get("hidden", False) is True:
            hide = "hidden"
            muted = " text-muted"
            checked = ""
            hidden_cols += 1

        data_attr = 'data-dmax="{}" data-dmin="{}" data-namespace="{}" {}'.format(
            header["dmax"], header["dmin"], header["namespace"], shared_key
        )

        ns = f'{header["namespace"]}: ' if header["namespace"] else ""
        cell_contents = f'<span class="mqc_table_tooltip" title="{ns}{header["description"]}">{header["title"]}</span>'

        t_headers[rid] = '<th id="header_{rid}" class="{rid} {h}" {da}>{c}</th>'.format(
            rid=rid, h=hide, da=data_attr, c=cell_contents
        )

        empty_cells[rid] = f'<td class="data-coloured {rid} {hide}"></td>'

        # Build the modal table row
        data = f"data-table-id='{dt.id}'"
        if violin_id:
            data += f" data-violin-id='{violin_id}'"
        t_modal_headers[rid] = f"""
        <tr class="{rid}{muted}" style="background-color: rgba({header["colour"]}, 0.15);">
          <td class="sorthandle ui-sortable-handle">||</span></td>
          <td style="text-align:center;">
            <input class="mqc_table_col_visible" type="checkbox" {checked} value="{rid}" {data}>
          </td>
          <td>{header["namespace"]}</td>
          <td>{header["title"]}</td>
          <td>{header["description"]}</td>
          <td><code>{k}</code></td>
          <td>{header.get("shared_key", "")}</td>
        </tr>"""

        # Make a colour scale
        if header["scale"] is False:
            c_scale = None
        else:
            c_scale = mqc_colour.mqc_colour_scale(
                name=header["scale"],
                minval=header["dmin"],
                maxval=header["dmax"],
                id=dt.id,
            )

        # Collect conditional formatting config
        cond_formatting_rules = {}
        if header.get("cond_formatting_rules"):
            cond_formatting_rules[rid] = header["cond_formatting_rules"]
        cond_formatting_rules.update(config.table_cond_formatting_rules)

        cond_formatting_colours = header.get("cond_formatting_colours", [])
        cond_formatting_colours.extend(config.table_cond_formatting_colours)

        # Add the data table cells
        for s_name, samp in dt.data[idx].items():
            if k in samp:
                val = samp[k]
                kname = f"{header['namespace']}_{rid}"
                dt.raw_vals[s_name][kname] = val

                if "modify" in header and callable(header["modify"]):
                    try:
                        val = header["modify"](val)
                    except TypeError as e:
                        logger.debug(f"Error modifying table value {kname} : {val} - {e}")

                if c_scale and c_scale.name not in c_scale.qualitative_scales:
                    try:
                        dmin = header["dmin"]
                        dmax = header["dmax"]
                        percentage = ((float(val) - dmin) / (dmax - dmin)) * 100
                        # Treat 0 as 0-width and make bars width of absolute value
                        if header.get("bars_zero_centrepoint"):
                            dmax = max(abs(header["dmin"]), abs(header["dmax"]))
                            dmin = 0
                            percentage = ((abs(float(val)) - dmin) / (dmax - dmin)) * 100
                        percentage = min(percentage, 100)
                        percentage = max(percentage, 0)
                    except (ZeroDivisionError, ValueError, TypeError):
                        percentage = 0
                else:
                    percentage = 100

                if "format" in header and callable(header["format"]):
                    valstring = header["format"](val)
                else:
                    try:
                        # "format" is a format string?
                        valstring = str(header["format"].format(val))
                    except ValueError:
                        try:
                            valstring = str(header["format"].format(float(val)))
                        except ValueError:
                            valstring = str(val)
                    except Exception:
                        valstring = str(val)

                    # This is horrible, but Python locale settings are worse
                    if config.thousandsSep_format is None:
                        config.thousandsSep_format = '<span class="mqc_small_space"></span>'
                    if config.decimalPoint_format is None:
                        config.decimalPoint_format = "."
                    valstring = valstring.replace(".", "DECIMAL").replace(",", "THOUSAND")
                    valstring = valstring.replace("DECIMAL", config.decimalPoint_format).replace(
                        "THOUSAND", config.thousandsSep_format
                    )

                suffix = header.get("suffix")
                if suffix:
                    # Add a space before the suffix, but not as an actual character, so ClipboardJS would copy
                    # the whole value without the space. Also, remove &nbsp; that we don't want ClipboardJS to copy.
                    suffix = suffix.replace("&nbsp;", " ").strip()
                    valstring += "<span class='mqc_small_space'></span>" + suffix

                # Conditional formatting
                # Build empty dict for cformatting matches
                cmatches = {}
                for cfc in cond_formatting_colours:
                    for cfck in cfc:
                        cmatches[cfck] = False
                # Find general rules followed by column-specific rules
                for cfk in ["all_columns", rid, dt.id]:
                    if cfk in cond_formatting_rules:
                        # Loop through match types
                        for ftype in cmatches.keys():
                            # Loop through array of comparison types
                            for cmp in cond_formatting_rules[cfk].get(ftype, []):
                                try:
                                    # Each comparison should be a dict with single key: val
                                    if "s_eq" in cmp and str(cmp["s_eq"]).lower() == str(val).lower():
                                        cmatches[ftype] = True
                                    if "s_contains" in cmp and str(cmp["s_contains"]).lower() in str(val).lower():
                                        cmatches[ftype] = True
                                    if "s_ne" in cmp and str(cmp["s_ne"]).lower() != str(val).lower():
                                        cmatches[ftype] = True
                                    if "eq" in cmp and float(cmp["eq"]) == float(val):
                                        cmatches[ftype] = True
                                    if "ne" in cmp and float(cmp["ne"]) != float(val):
                                        cmatches[ftype] = True
                                    if "gt" in cmp and float(cmp["gt"]) < float(val):
                                        cmatches[ftype] = True
                                    if "lt" in cmp and float(cmp["lt"]) > float(val):
                                        cmatches[ftype] = True
                                except Exception:
                                    logger.warning(f"Not able to apply table conditional formatting to '{val}' ({cmp})")
                # Apply HTML in order of config keys
                badge_col = None
                for cfc in cond_formatting_colours:
                    for cfck in cfc:  # should always be one, but you never know
                        if cmatches[cfck]:
                            badge_col = cfc[cfck]
                if badge_col is not None:
                    valstring = f'<span class="badge" style="background-color:{badge_col}">{valstring}</span>'

                # Determine background color based on scale. Only relevant for hashable values. If value is for some
                # reason a dict or a list, it's not hashable and the logic determining the color will not work.
                hashable = True
                try:
                    hash(val)
                except TypeError:
                    hashable = False
                    print(f"Value {val} is not hashable for table {dt.id}, column {k}, sample {s_name}")

                # Categorical background colours supplied
                if hashable and val in header.get("bgcols", {}).keys():
                    col = f"style=\"background-color:{header['bgcols'][val]} !important;\""
                    if s_name not in t_rows:
                        t_rows[s_name] = dict()
                    t_rows[s_name][rid] = f'<td val="{val}" class="{rid} {hide}" {col}>{valstring}</td>'

                # Build table cell background colour bar
                elif hashable and header["scale"]:
                    if c_scale is not None:
                        col = " background-color:{} !important;".format(
                            c_scale.get_colour(val, source=f'Table "{dt.id}", column "{k}"')
                        )
                    else:
                        col = ""
                    bar_html = f'<span class="bar" style="width:{percentage}%;{col}"></span>'
                    val_html = f'<span class="val">{valstring}</span>'
                    wrapper_html = f'<div class="wrapper">{bar_html}{val_html}</div>'

                    if s_name not in t_rows:
                        t_rows[s_name] = dict()
                    t_rows[s_name][rid] = f'<td val="{val}" class="data-coloured {rid} {hide}">{wrapper_html}</td>'

                # Scale / background colours are disabled
                else:
                    if s_name not in t_rows:
                        t_rows[s_name] = dict()
                    t_rows[s_name][rid] = f'<td val="{val}" class="{rid} {hide}">{valstring}</td>'

                # Is this cell hidden or empty?
                if s_name not in t_rows_empty:
                    t_rows_empty[s_name] = dict()
                t_rows_empty[s_name][rid] = header.get("hidden", False) or str(val).strip() == ""

        # Remove header if we don't have any filled cells for it
        if sum([len(rows) for rows in t_rows.values()]) == 0:
            if header.get("hidden", False) is True:
                hidden_cols -= 1
            t_headers.pop(rid, None)
            t_modal_headers.pop(rid, None)
            logger.debug(f"Removing header {k} from table, as no data")

    #
    # Put everything together
    #

    # Buttons above the table
    html = ""
    if not config.simple_output:
        # Copy Table Button
        buttons = []

        buttons.append(
            f"""
        <button type="button" class="mqc_table_copy_btn btn btn-default btn-sm" data-clipboard-target="table#{dt.id}">
            <span class="glyphicon glyphicon-copy"></span> Copy table
        </button>
        """
        )

        # Configure Columns Button
        if len(t_headers) > 1:
            buttons.append(
                f"""
            <button type="button" class="mqc_table_configModal_btn btn btn-default btn-sm" data-toggle="modal" 
                data-target="#{dt.id}_configModal">
                <span class="glyphicon glyphicon-th"></span> Configure columns
            </button>
            """
            )

        # Sort By Highlight button
        buttons.append(
            f"""
        <button type="button" class="mqc_table_sortHighlight btn btn-default btn-sm" 
            data-target="#{dt.id}" data-direction="desc" style="display:none;">
            <span class="glyphicon glyphicon-sort-by-attributes-alt"></span> Sort by highlight
        </button>
        """
        )

        # Scatter Plot Button
        if len(t_headers) > 1:
            buttons.append(
                f"""
            <button type="button" class="mqc_table_makeScatter btn btn-default btn-sm" 
                data-toggle="modal" data-target="#tableScatterModal" data-table="#{dt.id}">
                <span class="glyphicon glyphicon glyphicon-equalizer"></span> Scatter plot
            </button>
            """
            )

        if violin_id is not None:
            buttons.append(
                f"""
            <button type="button" class="mqc-table-to-violin btn btn-default btn-sm" 
                data-table-id="{dt.id}" data-violin-id="{violin_id}">
                <span class="glyphicon glyphicon-align-left"></span> Violin plot
            </button>
            """
            )

        buttons.append(
            f"""
        <button type="button" class="export-plot btn btn-default btn-sm" 
            data-pid="{violin_id or dt.id}" data-type="table"
        >Export as CSV</button>
        """
        )

        # "Showing x of y columns" text
        row_visibilities = [all(t_rows_empty[s_name].values()) for s_name in t_rows_empty]
        visible_rows = [x for x in row_visibilities if not x]

        # Visible rows
        t_showing_rows_txt = f'Showing <sup id="{dt.id}_numrows" class="mqc_table_numrows">{len(visible_rows)}</sup>/<sub>{len(t_rows)}</sub> rows'

        # How many columns are visible?
        ncols_vis = (len(t_headers) + 1) - hidden_cols
        t_showing_cols_txt = ""
        if len(t_headers) > 1:
            t_showing_cols_txt = f' and <sup id="{dt.id}_numcols" class="mqc_table_numcols">{ncols_vis}</sup>/<sub>{len(t_headers)}</sub> columns'

        # Build table header text
        buttons.append(
            f"""
        <small id="{dt.id}_numrows_text" class="mqc_table_numrows_text">{t_showing_rows_txt}{t_showing_cols_txt}.</small>
        """
        )

        panel = "\n".join(buttons)
        html += f"""
        <div class='row'>\n<div class='col-xs-12'>\n{panel}\n</div>\n</div>
        """

    # Build the table itself
    collapse_class = "mqc-table-collapse" if len(t_rows) > 10 and config.collapse_tables else ""
    html += f"""
        <div id="{dt.id}_container" class="mqc_table_container">
            <div class="table-responsive mqc-table-responsive {collapse_class}">
                <table id="{dt.id}" class="table table-condensed mqc_table" data-title="{table_title}" data-sortlist="{_get_sortlist(dt)}">
        """

    # Build the header row
    col1_header = dt.pconfig.get("col1_header", "Sample Name")
    html += f"<thead><tr><th class=\"rowheader\">{col1_header}</th>{''.join(t_headers.values())}</tr></thead>"

    # Build the table body
    html += "<tbody>"
    t_row_keys = t_rows.keys()
    if dt.pconfig.get("sortRows") is not False:
        t_row_keys = sorted(t_row_keys)
    for s_name in t_row_keys:
        # Hide the row if all cells are empty or hidden
        row_hidden = ' style="display:none"' if all(t_rows_empty[s_name].values()) else ""
        html += f"<tr{row_hidden}>"
        # Sample name row header
        html += f'<th class="rowheader" data-original-sn="{s_name}">{s_name}</th>'
        for k in t_headers:
            html += t_rows[s_name].get(k, empty_cells[k])
        html += "</tr>"
    html += "</tbody></table></div>"
    if len(t_rows) > 10 and config.collapse_tables:
        html += '<div class="mqc-table-expand"><span class="glyphicon glyphicon-chevron-down" aria-hidden="true"></span></div>'
    html += "</div>"

    # Save the raw values to a file if requested
    if dt.pconfig.get("save_file") is True:
        fn = dt.pconfig.get("raw_data_fn", f"multiqc_{dt.id}")
        util_functions.write_data_file(dt.raw_vals, fn)
        report.saved_raw_data[fn] = dt.raw_vals

    # Build the bootstrap modal to customise columns and order
    modal = ""
    if not config.simple_output:
        modal = _configuration_modal(
            tid=dt.id,
            title=table_title,
            trows="".join(t_modal_headers.values()),
            violin_id=violin_id,
        )

    return html, modal


def _configuration_modal(tid: str, title: str, trows: str, violin_id: Optional[str] = None) -> str:
    data = f"data-table-id='{tid}'"
    if violin_id is not None:
        data += f" data-violin-id='{violin_id}'"
    return f"""
    <!-- MultiQC Table Columns Modal -->
    <div class="modal fade" id="{tid}_configModal" tabindex="-1">
      <div class="modal-dialog modal-lg">
        <div class="modal-content">
          <div class="modal-header">
            <button type="button" class="close" data-dismiss="modal" aria-label="Close"><span aria-hidden="true">&times;</span></button>
            <h4 class="modal-title">{title}: Columns</h4>
          </div>
          <div class="modal-body">
            <p>Uncheck the tick box to hide columns. Click and drag the handle on the left to change order. Table ID: <code>{tid}</code></p>
            <p>
                <button class="btn btn-default btn-sm mqc_configModal_bulkVisible" {data} data-action="showAll">Show All</button>
                <button class="btn btn-default btn-sm mqc_configModal_bulkVisible" {data} data-action="showNone">Show None</button>
            </p>
            <table class="table mqc_table mqc_sortable mqc_configModal_table" id="{tid}_configModal_table" data-title="{title}">
              <thead>
                <tr>
                  <th class="sorthandle" style="text-align:center;">Sort</th>
                  <th style="text-align:center;">Visible</th>
                  <th>Group</th>
                  <th>Column</th>
                  <th>Description</th>
                  <th>ID</th>
                  <th>Scale</th>
                </tr>
              </thead>
              <tbody>
                {trows}
              </tbody>
            </table>
        </div>
        <div class="modal-footer"> <button type="button" class="btn btn-default" data-dismiss="modal">Close</button> </div>
    </div> </div> </div>"""


def _get_sortlist(dt: DataTable) -> str:
    """
    Custom column sorting order for a table plot. The order is provided in the following form:

    ```yaml
    custom_plot_config:
      general_stats_table:
        defaultsort:
          - column: "Mean Insert Length"
            direction: asc
          - column: "Starting Amount (ng)"
      quast_table:
        defaultsort:
        - column: "Largest contig"
    ```

    It is returned in a form os a list literal, as expected by the jQuery tablesorter plugin.
    """
    defaultsort = dt.pconfig.get("defaultsort")
    if defaultsort is None:
        return ""

    headers = dt.get_headers_in_order()
    sortlist = []

    # defaultsort is a list of {column, direction} objects
    for d in defaultsort:
        try:
            # The first element of the triple is not actually unique, it's a bucket index,
            # so we must re-enumerate ourselves here
            idx = next(
                idx
                for idx, (_, k, header) in enumerate(headers)
                if d["column"].lower() in [k.lower(), header["title"].lower()]
            )
        except StopIteration:
            logger.warning(
                "Tried to sort by column '%s', but column was not found. Available columns: %s",
                d["column"],
                [k for (_, k, _) in headers],
            )
            return ""
        idx += 1  # to account for col1_header
        direction = 0 if d.get("direction", "").startswith("asc") else 1
        sortlist.append([idx, direction])

    return str(sortlist)



import dataclasses
import io
import logging
import os
from typing import Dict, List, Union, Optional
import plotly.graph_objects as go

from multiqc.plots.plotly.plot import Plot, PlotType, BaseDataset
from multiqc.utils import util_functions, config
from multiqc.utils.config import update_dict

logger = logging.getLogger(__name__)


# {"name": "SAMPLE1", "color": "#111111", "data": [[x, y], [x, y], ...]}
LineT = Dict[str, Union[str, List[List[float]]]]


def plot(lists_of_lines: List[List[LineT]], pconfig: Dict) -> str:
    """
    Build and add the plot data to the report, return an HTML wrapper.
    :param lists_of_lines: each dataset is a 2D dict, first keys as sample names, then x:y data pairs
    :param pconfig: dict with config key:value pairs. See CONTRIBUTING.md
    :return: HTML with JS, ready to be inserted into the page
    """

    # if self.n_samples >= config.max_table_rows:
    # Get a line of median values at each point with an interval of max and
    # min values
    # Create a violin of median values in each sample, showing dots for outliers
    # Clicking on a dot of a violin will show the line plot for that sample

    p = LinePlot(pconfig, lists_of_lines)

    from multiqc.utils import report

    return p.add_to_report(report)


class LinePlot(Plot):
    @dataclasses.dataclass
    class Dataset(BaseDataset):
        lines: List[Dict]

        @staticmethod
        def create(
            dataset: BaseDataset,
            lines: List[Dict],
            pconfig: Dict,
        ) -> "LinePlot.Dataset":
            dataset = LinePlot.Dataset(
                **dataset.__dict__,
                lines=lines,
            )
            dataset.dconfig["categories"] = dataset.dconfig.get("categories", pconfig.get("categories", []))
            if dataset.dconfig["categories"]:
                # Prevent JavaScript from automatically parsing categorical values as numbers
                dataset.layout["xaxis"]["type"] = "category"
                # check that all lines have the same number of categories
                assert all(
                    len(line["data"]) == len(dataset.dconfig["categories"]) for line in dataset.lines
                ), dataset.uid

            # convert HighCharts hardcoded trace parameters to Plotly
            lines = []
            for src_line in dataset.lines:
                new_line = {
                    "name": src_line["name"],
                    "data": src_line["data"],
                    "color": src_line.get("color"),
                }
                lines.append(new_line)
                new_line["marker"] = {}
                if "dashStyle" in src_line:
                    new_line["line"] = new_line.get("line", {})
                    new_line["line"]["dash"] = convert_dash_style(src_line["dashStyle"])
                if "lineWidth" in src_line:
                    new_line["line"] = new_line.get("line", {})
                    new_line["line"]["width"] = src_line["lineWidth"]
                if "showInLegend" in src_line:
                    new_line["showlegend"] = src_line["showInLegend"]
                if "marker" in src_line:
                    new_line["marker"] = new_line.get("marker", {})
                    new_line["marker"]["line"] = new_line["marker"].get("line", {})
                    if "lineWidth" in src_line["marker"]:
                        new_line["marker"]["line"]["width"] = src_line["marker"]["lineWidth"]
                    if "lineColor" in src_line["marker"]:
                        new_line["marker"]["line"]["color"] = src_line["marker"]["lineColor"]
                    if "symbol" in src_line["marker"]:
                        new_line["mode"] = "lines+markers"
                        new_line["marker"]["symbol"] = src_line["marker"]["symbol"]
            dataset.lines = lines
            # Update default trace parameters
            dataset.trace_params.update(
                mode="lines" if config.lineplot_style == "lines" else "lines+markers",
                line={"width": 2 if config.lineplot_style == "lines" else 0.6},
                marker={"size": 4},
            )
            return dataset

        def create_figure(
            self,
            layout: go.Layout,
            is_log=False,
            is_pct=False,
        ) -> go.Figure:
            """
            Create a Plotly figure for a dataset
            """
            if self.plot.flat:
                layout.height += len(self.lines) * 5  # extra space for legend

            fig = go.Figure(layout=layout)
            for line in self.lines:
                if len(line["data"]) > 0 and isinstance(line["data"][0], list):
                    xs = [x[0] for x in line["data"]]
                    ys = [x[1] for x in line["data"]]
                elif self.dconfig.get("categories"):
                    assert len(line["data"]) == len(self.dconfig["categories"])
                    xs = self.dconfig["categories"]
                    ys = line["data"]
                else:
                    xs = [x for x in range(len(line["data"]))]
                    ys = line["data"]

                params = dict(
                    marker=line.get("marker", {}),
                    line=line.get("line", {}),
                    showlegend=line.get("showlegend", None),
                    mode=line.get("mode", None),
                )
                params = update_dict(params, self.trace_params, none_only=True)
                params["marker"]["color"] = line.get("color")

                fig.add_trace(
                    go.Scatter(
                        x=xs,
                        y=ys,
                        name=line["name"],
                        text=[line["name"]] * len(xs),
                        **params,
                    )
                )
            return fig

    def __init__(self, pconfig: Dict, lists_of_lines: List[List[LineT]]):
        super().__init__(PlotType.LINE, pconfig, len(lists_of_lines))

        self.datasets: List[LinePlot.Dataset] = [
            LinePlot.Dataset.create(d, lines, pconfig) for d, lines in zip(self.datasets, lists_of_lines)
        ]

        # Make a tooltip always show on hover over any point on plot
        self.layout.hoverdistance = -1

        y_min_range = pconfig.get("yMinRange")
        y_bands = pconfig.get("yPlotBands")
        x_bands = pconfig.get("xPlotBands")
        x_lines = pconfig.get("xPlotLines")
        y_lines = pconfig.get("yPlotLines")
        if y_min_range or y_bands or y_lines:
            # We don't want the bands to affect the calculated axis range, so we
            # find the min and the max from data points, and manually set the range
            for dataset in self.datasets:
                ymin = dataset.layout["yaxis"]["autorangeoptions"]["clipmin"]
                if ymin is None:
                    ymin = dataset.layout["yaxis"]["autorangeoptions"]["minallowed"]
                ymax = dataset.layout["yaxis"]["autorangeoptions"]["clipmax"]
                if ymax is None:
                    ymax = dataset.layout["yaxis"]["autorangeoptions"]["maxallowed"]
                if ymin is None or ymax is None:
                    for line in dataset.lines:
                        if len(line["data"]) > 0 and isinstance(line["data"][0], list):
                            ys = [x[1] for x in line["data"]]
                        else:
                            ys = line["data"]
                        if ymin is None:
                            ymin = min(ys)
                        if ymax is None:
                            ymax = max(ys)
                            ymax += (ymax - ymin) * 0.05
                dataset.layout["yaxis"]["range"] = [ymin, ymax]

        if x_bands or x_lines:
            # same as above but for x-axis
            for dataset in self.datasets:
                xmin = dataset.layout["xaxis"]["autorangeoptions"]["clipmin"]
                if xmin is None:
                    xmin = dataset.layout["xaxis"]["autorangeoptions"]["minallowed"]
                xmax = dataset.layout["xaxis"]["autorangeoptions"]["clipmax"]
                if xmax is None:
                    xmax = dataset.layout["xaxis"]["autorangeoptions"]["maxallowed"]
                if xmin is None or xmax is None:
                    for line in dataset.lines:
                        if len(line["data"]) > 0 and isinstance(line["data"][0], list):
                            xs = [x[0] for x in line["data"]]
                        else:
                            xs = [x for x in range(len(line["data"]))]
                        if xmin is None:
                            xmin = min(xs)
                        if xmax is None:
                            xmax = max(xs)
                dataset.layout["xaxis"]["range"] = [xmin, xmax]

        self.layout.shapes = (
            [
                dict(
                    type="rect",
                    y0=band["from"],
                    y1=band["to"],
                    x0=0,
                    x1=1,
                    xref="paper",  # make x coords are relative to the plot paper [0,1]
                    fillcolor=band["color"],
                    line={
                        "width": 0,
                    },
                    layer="below",
                )
                for band in (y_bands or [])
            ]
            + [
                dict(
                    type="rect",
                    x0=band["from"],
                    x1=band["to"],
                    y0=0,
                    y1=1,
                    yref="paper",  # make y coords are relative to the plot paper [0,1]
                    fillcolor=band["color"],
                    line={
                        "width": 0,
                    },
                    layer="below",
                )
                for band in (x_bands or [])
            ]
            + [
                dict(
                    type="line",
                    xref="paper",
                    yref="y",
                    x0=0,
                    y0=line["value"],
                    x1=1,
                    y1=line["value"],
                    line={
                        "width": line["width"],
                        "dash": convert_dash_style(line.get("dashStyle")),
                        "color": line["color"],
                    },
                )
                for line in (y_lines or [])
            ]
            + [
                dict(
                    type="line",
                    yref="paper",
                    xref="x",
                    x0=line["value"],
                    y0=0,
                    x1=line["value"],
                    y1=1,
                    line={
                        "width": line["width"],
                        "dash": convert_dash_style(line.get("dashStyle")),
                        "color": line["color"],
                    },
                )
                for line in (x_lines or [])
            ]
        )

    @staticmethod
    def tt_label() -> Optional[str]:
        """Default tooltip label"""
        return "<br>%{x}: %{y}"

    def axis_controlled_by_switches(self) -> List[str]:
        """
        Return a list of axis names that are controlled by the log10 scale and percentage
        switch buttons
        """
        return ["yaxis"]

    def save_data_file(self, dataset: Dataset) -> None:
        y_by_x_by_sample = dict()
        last_cats = None
        shared_cats = True
        for line in dataset.lines:
            y_by_x_by_sample[line["name"]] = dict()

            # Check to see if all categories are the same
            if len(line["data"]) > 0 and isinstance(line["data"][0], list):
                if last_cats is None:
                    last_cats = [x[0] for x in line["data"]]
                elif last_cats != [x[0] for x in line["data"]]:
                    shared_cats = False

            for i, x in enumerate(line["data"]):
                if isinstance(x, list):
                    y_by_x_by_sample[line["name"]][x[0]] = x[1]
                else:
                    try:
                        y_by_x_by_sample[line["name"]][dataset.dconfig["categories"][i]] = x
                    except (ValueError, KeyError, IndexError):
                        y_by_x_by_sample[line["name"]][str(i)] = x

        # Custom tsv output if the x-axis varies
        if not shared_cats and config.data_format in ["tsv", "csv"]:
            sep = "\t" if config.data_format == "tsv" else ","
            fout = ""
            for line in dataset.lines:
                fout += line["name"] + sep + "X" + sep + sep.join([str(x[0]) for x in line["data"]]) + "\n"
                fout += line["name"] + sep + "Y" + sep + sep.join([str(x[1]) for x in line["data"]]) + "\n"

            fn = f"{dataset.uid}.{config.data_format_extensions[config.data_format]}"
            fpath = os.path.join(config.data_dir, fn)
            with io.open(fpath, "w", encoding="utf-8") as f:
                f.write(fout.encode("utf-8", "ignore").decode("utf-8"))
        else:
            util_functions.write_data_file(y_by_x_by_sample, dataset.uid)


def convert_dash_style(dash_style: str) -> str:
    """Convert dash style from Highcharts to Plotly"""
    return {
        "Solid": "solid",
        "ShortDash": "dash",
        "ShortDot": "dot",
        "ShortDashDot": "dashdot",
        "ShortDashDotDot": "dashdot",
        "Dot": "dot",
        "Dash": "dash",
        "DashDot": "dashdot",
        "LongDash": "longdash",
        "LongDashDot": "longdashdot",
        "LongDashDotDot": "longdashdot",
    }.get(dash_style, "solid")



import logging
import os
import re
from typing import Dict, List, Optional, Union

from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


def read_histogram(module, program_key, headers, formats, picard_tool, sentieon_algo=None):
    """
    Reads a Picard HISTOGRAM file.

    Args:
        module: the Picard QC module
        program_key: the key used to find the program (ex. picard/quality_by_cycle)
        headers: the list of expected headers for the histogram
        formats: the list of methods to apply to re-format each field (on a given row)
        picard_tool: the name of the Picard tool to be found in the header, e.g. MeanQualityByCycle
        sentieon_algo: the name of the Sentieon algorithm to be found in the header, e.g. MeanQualityByCycle
    """
    all_data = dict()
    assert len(formats) == len(headers)
    sample_data = None

    # Go through logs and find Metrics
    for f in module.find_log_files(program_key, filehandles=True):
        s_name = f["s_name"]
        for line in f["f"]:
            maybe_s_name = extract_sample_name(
                module,
                line,
                f,
                picard_tool=picard_tool,
                sentieon_algo=sentieon_algo,
            )
            if maybe_s_name:
                s_name = maybe_s_name
                sample_data = None

            if is_line_right_before_table(line, sentieon_algo=sentieon_algo):
                # check the header
                line = f["f"].readline()
                if line.strip().split("\t") == headers:
                    sample_data = dict()
                else:
                    sample_data = None

            elif sample_data is not None:
                fields = line.strip().split("\t")
                if len(fields) == len(headers):
                    for i in range(len(fields)):
                        fields[i] = formats[i](fields[i])
                    sample_data[fields[0]] = dict(zip(headers, fields))

        # append the data
        if sample_data:
            if s_name in all_data:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
            all_data[s_name] = sample_data

            module.add_data_source(f, s_name, section="Histogram")
            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            module.add_software_version(None, s_name)

    data = module.ignore_samples(all_data)

    # Write data to file
    module.write_data_file(data, f"{module.anchor}_histogram")

    return data


def is_line_right_before_table(
    line: str,
    picard_class: Union[None, str, List[str]] = None,
    sentieon_algo: Optional[str] = None,
) -> bool:
    """
    Picard logs from different samples can be concatenated together, so the module
    needs to know a marker to find where new sample information starts.

    The command line Picard tools themselves use the "## METRICS CLASS" header
    line for that purpose; however, the Picard classes often used by other
    tools and platforms - e.g. Sentieon and Parabricks  - while adding their own
    headers, so we need to handle them as well.
    """
    if isinstance(picard_class, list):
        picard_classes = picard_class
    elif picard_class is None:
        picard_classes = []
    else:
        picard_classes = [picard_class]
    return (
        (line.startswith("## METRICS CLASS") or line.startswith("## HISTOGRAM"))
        and (not picard_classes or any(c.upper() in line.upper() for c in picard_classes))
        or sentieon_algo
        and line.startswith("#SentieonCommandLine:")
        and f" --algo {sentieon_algo}" in line
    )


def extract_sample_name(
    mod,
    line: str,
    f: Dict,
    picard_tool: Union[str, List[str]],
    sentieon_algo: Optional[str] = None,
    picard_opt: Union[None, str, List[str]] = None,
) -> Optional[str]:
    """
    Historically, MultiQC supported Picard tools QC outputs merged together into
    one file from different samples and tools. In order to handle this correctly,
    MultiQC needs to take the sample name elsewhere rather than the file name,
    so it tries to parse the command line recorded in the output header.

    This approach can be disabled with `config.picard_config.s_name_filenames`
    """
    if getattr(config, "picard_config", {}).get("s_name_filenames", False):
        return None

    # Name of the option that contains the name of the input file to fetch the sample name.
    # Examples of the commands:
    # CollectIlluminaLaneMetrics --OUTPUT_PREFIX 200908_J00178_0673_AHJ5FLBBXY
    # picard.analysis.CollectAlignmentSummaryMetrics INPUT=/160219_0463_1_ACTTGA_GL_020516_SpHn_RNF212.mm10.bam_sorted
    picard_opt = picard_opt or ["INPUT"]
    picard_opts = picard_opt if isinstance(picard_opt, list) else [picard_opt]
    picard_tools = picard_tool if isinstance(picard_tool, list) else [picard_tool]

    picard_command = (
        line.startswith("# ")
        and any(pt.upper() in line.upper() for pt in picard_tools)
        and any(
            f" {po}=" in line.upper()
            or f" --{po}=" in line.upper()
            or f"{po}" in line.upper().split()
            or f"--{po}" in line.upper().split()
            for po in picard_opts
        )
    )
    sentieon_command = (
        sentieon_algo
        and line.startswith("#SentieonCommandLine:")
        and f" --algo {sentieon_algo}" in line
        and " -i " in line
    )
    # Pull sample name from the input file name, recorded in the command line:
    match = None
    if picard_command:
        for po in picard_opts:
            match = re.search(rf"{po}(?:=|\s+)(\[?[^\s]+\]?)", line, flags=re.IGNORECASE)
            if match is not None:
                break
    elif sentieon_command:
        match = re.search(r" -i\s+(\[?\S+\]?)", line, flags=re.IGNORECASE)
    if match:
        f_name = os.path.basename(match.group(1).strip("[]"))
        s_name = mod.clean_s_name(f_name, f)
        return s_name
    return None


def multiply_hundred(val):
    try:
        val = float(val) * 100
    except ValueError:
        pass
    return val



""" MultiQC module to parse output from nonpareil """


import logging
import numpy as np

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.utils import mqc_colour


# Initialise the logger
from multiqc.plots import table, linegraph

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Nonpareil analysis is split into two parts: the first (written in C++)
    performs the subsampling, while the second (written in R) performs the
    statistical analyses. As such, this model requires the user to post-process
    the R object from the second part."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="nonpareil",
            anchor="nonpareil",
            href="https://github.com/lmrodriguezr/nonpareil",
            info="Estimate metagenomic coverage and sequence diversity ",
            doi="10.1093/bioinformatics/btt584",
        )
        # Config options
        self.plot_colours = getattr(config, "nonpareil", {}).get("plot_colours", "Paired")

        # Read JSON file
        self.data_by_sample = dict()
        for f in self.find_log_files("nonpareil", filehandles=True):
            json_parsed = self.parse_nonpareil_json(f)
            common_samples = set(json_parsed).intersection(self.data_by_sample)
            if common_samples:
                log.debug(f"Duplicate sample names found ({common_samples})! Overwriting...")
            self.data_by_sample.update(json_parsed)
            for s_name in json_parsed:
                self.add_data_source(f, s_name)

        # Filter to strip out ignored sample names
        self.data_by_sample = self.ignore_samples(self.data_by_sample)
        if len(self.data_by_sample) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.data_by_sample)} reports")
        self.write_data_file(self.data_by_sample, "nonpareil")

        # Add versions
        for s_name, data in self.data_by_sample.items():
            self.add_software_version(str(data["nonpareil_version"]), s_name)

        # Add general stats
        self.nonpareil_general_stats_table()
        # Add table
        self.nonpareil_stats_table()
        # Add plots
        self.nonpareil_redundancy_plot()

    def parse_nonpareil_json(self, f):
        """Parse the JSON output from nonpareil"""
        import json

        try:
            json_raw = json.load(f["f"])
        except json.JSONDecodeError as e:
            log.warning(f"Could not parse nonpareil JSON: '{f['fn']}': {e}, skipping file")
            return None, {}

        # Rename samples, if label available
        for s_name in list(json_raw.keys()):
            s_label = json_raw[s_name].get("label")
            if s_label:
                json_raw[s_label] = json_raw.pop(s_name)

        for s_name, data in json_raw.items():
            # Convert base pairs to megabase pairs
            data["x.adj"] = [(1e-6 if x == 0 else x) / 1000 / 1000 for x in data["x.adj"]]
            if data["has.model"]:
                data["x.model"] = [x / 1000 / 1000 for x in data["x.model"]]
            # Convert fraction to percentage
            data["y.cov"] = [x * 100 for x in data["y.cov"]]
            if data["has.model"]:
                data["y.model"] = [x * 100 for x in data["y.model"]]
            data["y.sd"] = [x * 100 for x in data["y.sd"]]
            data["y.p75"] = [x * 100 for x in data["y.p75"]]
            data["y.p25"] = [x * 100 for x in data["y.p25"]]
            # Prepare plot data
            data["observed"] = {x: y for x, y in zip(data["x.adj"], data["y.cov"])}
            if data["has.model"]:
                data["model"] = {x: y for x, y in zip(data["x.model"], data["y.model"])}
            else:
                modelR = data["modelR"]
                assert isinstance(modelR, list) and len(modelR) == 0, "there is no model, but modelR is not empty"
                data["modelR"] = np.nan
                LRstar = data["LRstar"]
                assert isinstance(LRstar, list) and len(LRstar) == 0, "there is no model, but LRstar is not empty"
                data["LRstar"] = np.nan
            # Add prefix to labels
            for key in list(data.keys()):
                data["nonpareil_" + key] = data.pop(key)

        return json_raw

    def nonpareil_general_stats_table(self):
        """Take the parsed stats from the nonpareil report and add it to the
        General Statistics table at the top of the report"""

        headers = {
            "nonpareil_R": {
                "title": f"{config.read_count_prefix} Reads",
                "description": f"Total raw sequences ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "min": 0,
                "scale": "RdYlGn",
                "shared_key": "read_count",
                "hidden": True,
            },
            "nonpareil_LR": {
                "title": f"{config.base_count_prefix} Seq. effort",
                "description": f"Total base pairs sequenced ({config.base_count_desc})",
                "modify": lambda x: x * config.base_count_multiplier,
                "min": 0,
                "scale": "Blues",
                "shared_key": "base_count",
                "hidden": True,
            },
            "nonpareil_kappa": {
                "title": "Redundancy",
                "description": "Dataset redundancy",
                "modify": lambda x: x * 100,
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn-rev",
            },
            "nonpareil_C": {
                "title": "Coverage",
                "description": "Dataset coverage",
                "modify": lambda x: x * 100,
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "nonpareil_diversity": {
                "title": "Diversity",
                "description": "Dataset Nd index of sequence diversity",
                "min": 0,
                "scale": "GnBu-rev",
            },
        }

        self.general_stats_addcols(self.data_by_sample, headers)

    def nonpareil_stats_table(self):
        """Take the parsed stats from the nonpareil report and add it to the
        General Statistics table at the top of the report"""

        headers = {
            "nonpareil_L": {
                "title": "Read Length",
                "min": 0,
                "suffix": " bp",
                "scale": "GnBu",
                "format": "{:,.3f}",
                "hidden": True,
            },
            "nonpareil_AL": {
                "title": "Adjusted read length",
                "description": "Adjusted read length (if using kmers - same as read length for alignment)",
                "min": 0,
                "scale": "GnBu",
                "suffix": " bp",
                "format": "{:,.3f}",
                "hidden": True,
            },
            "nonpareil_R": {
                "title": f"{config.read_count_prefix} Reads",
                "description": f"Total raw sequences ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "min": 0,
                "scale": "RdYlGn",
                "shared_key": "read_count",
            },
            "nonpareil_LR": {
                "title": f"{config.base_count_prefix} Seq. effort",
                "description": f"Total base pairs sequenced ({config.base_count_desc})",
                "modify": lambda x: x * config.base_count_multiplier,
                "min": 0,
                "scale": "Blues",
                "shared_key": "base_count",
            },
            "nonpareil_overlap": {
                "title": "Read overlap",
                "description": "Minimum read overlap",
                "min": 0,
                "suffix": " bp",
                "hidden": True,
                "format": "{:,.0f}",
            },
            "nonpareil_log.sample": {
                "title": "Log sampling",
                "description": "Multiplier of the log-sampling (or zero if linear)",
                "max": 1,
                "min": 0,
                "hidden": True,
            },
            "nonpareil_kappa": {
                "title": "Redundancy",
                "description": "Dataset redundancy",
                "modify": lambda x: x * 100,
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn-rev",
            },
            "nonpareil_C": {
                "title": "Coverage",
                "description": "Dataset coverage",
                "modify": lambda x: x * 100,
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "nonpareil_consistent": {
                "title": "Consistent Data?",
                "description": "Is the data sufficient for accurate estimation?",
                "modify": lambda x: "Yes" if x == 1 else "No",
                "scale": False,
            },
            "nonpareil_star": {
                "title": "Ideal coverage",
                "description": "Objective coverage in percentage; i.e., coverage value considered near-complete.",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": False,
                "hidden": True,
            },
            "nonpareil_has.model": {
                "title": "Model available?",
                "description": "Was model estimated?",
                "modify": lambda x: "Yes" if x else "No",
                "scale": False,
                "hidden": True,
            },
            "nonpareil_LRstar": {
                "title": f"{config.base_count_prefix} Ideal seq. effort",
                "description": f"Projected sequencing effort for nearly complete coverage ({config.base_count_desc})",
                "modify": lambda x: x * config.base_count_multiplier,
                "min": 0,
                "scale": "RdYlGn-rev",
                "shared_key": "base_count",
            },
            "nonpareil_modelR": {
                "title": "Model correlation",
                "description": "Pearson R for the estimated model",
                "max": 1,
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.3f}",
                "hidden": True,
            },
            "nonpareil_diversity": {
                "title": "Diversity",
                "description": "Dataset Nd index of sequence diversity",
                "min": 0,
                "scale": "GnBu-rev",
            },
        }

        pconfig = {
            "id": "nonpareil-table",
            "title": "Nonpareil: statistics",
        }

        self.add_section(
            name="Statistics",
            anchor="nonpareil-table-section",
            description="""
            Nonpareil uses the redundancy of the reads in metagenomic datasets to
            estimate the average coverage and predict the amount of sequences that
            will be required to achieve "nearly complete coverage".
            """,
            helptext="""
            From the Nonpareil [FAQ](http://enve-omics.ce.gatech.edu/nonpareil/faq):

            **What is the Nonpareil diversity index?**

            We haven't yet published a formal description of the "diversity" reported by Nonpareil, which we internally refer to as Nonpareil diversity. The Nonpareil diversity is a measurement of how complex a community is in "sequence space". Graphically, the index is a measurement of "how much to the right the Nonpareil curve is".

            **What's the expected range of the Nonpareil diversity index?**

            We've found that the Nonpareil diversity index ranks communities very
            reliably in terms of diversity, and can even detect small seasonal
            variations in bacterial communities.
            For reference: an environment largely dominated by a single bacterial
            species like human posterior fornix or the Acid Mine Drainage gets
            values around 15 to 16, human stool samples around 17 to 19, freshwater
            and sandy soils around 20 to 22, and marine and other soils around
            21 to 25. This index has a logarithmic scale, to allow meaningful
            comparisons between extreme cases, like single species vs soil.
            """,
            plot=table.plot(self.data_by_sample, headers, pconfig),
        )

    def nonpareil_redundancy_plot(self):
        """Make the redundancy plot for nonpareil"""

        esconfig = {
            "dashStyle": "Dash",
            "lineWidth": 2,
            "marker": {"enabled": False},
            "enableMouseTracking": True,
            "showInLegend": False,
        }

        data_colors_default = mqc_colour.mqc_colour_scale().get_colours(self.plot_colours)
        data_colors = {
            s_name: data.get("nonpareil_col", data_colors_default.pop(0))
            for s_name, data in self.data_by_sample.items()
        }

        data_labels = [
            {"name": "Combined"},
            {"name": "Observed"},
            {"name": "Model"},
        ]
        data_plot = list()
        extra_series = list()
        for idx, dataset in enumerate(data_labels):
            data_plot.append(dict())
            extra_series.append(list())
            for s_name, data in self.data_by_sample.items():
                if dataset["name"] == "Observed":
                    data_plot[idx][s_name] = data["nonpareil_observed"]
                elif dataset["name"] == "Model" and data["nonpareil_has.model"]:
                    data_plot[idx][s_name] = data["nonpareil_model"]
                elif dataset["name"] == "Combined":
                    data_plot[idx][s_name] = data["nonpareil_observed"]
                    if data["nonpareil_has.model"]:
                        extra_series[idx].append(dict(esconfig))
                        extra_series[idx][-1]["name"] = s_name
                        extra_series[idx][-1]["data"] = [[x, y] for x, y in data["nonpareil_model"].items()]
                        extra_series[idx][-1]["color"] = data_colors[s_name]

        pconfig = {
            "id": "nonpareil-redundancy-plot",
            "colors": data_colors,
            "title": "Nonpareil: Redundancy levels",
            "xlab": "Sequencing effort (Mbp)",
            "ylab": "Estimated Average Coverage",
            "xmin": 1e-3,
            "ymin": 0,
            "ymax": 100,
            "xDecimals": True,
            "yDecimals": True,
            "xLog": True,
            "tt_label": "{point.x:.2f} Mbp: {point.y:.2f}",
            "extra_series": extra_series,
            "data_labels": data_labels,
        }

        self.add_section(
            name="Redundancy levels",
            anchor="nonpareil-redundancy",
            description="Observed and modelled redundancy levels across samples.",
            helptext="""
            The estimation of the Redundancy is at the core of Nonpareil,
            but it's when those values are transformed into average coverage
            that they become comparable across samples, and become useful for
            project design and sample evaluation.
            """,
            plot=linegraph.plot(data_plot, pconfig),
        )



import json
import logging
import os
from collections import defaultdict
from itertools import islice

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="bcl2fastq",
            anchor="bcl2fastq",
            href="https://support.illumina.com/sequencing/sequencing_software/bcl2fastq-conversion-software.html",
            info="can be used to both demultiplex data and convert BCL files to FASTQ file formats for downstream analysis.",
            # Can't find a DOI // doi=
        )

        # Gather data from all json files
        self.bcl2fastq_data = dict()
        for f in self.find_log_files("bcl2fastq"):
            self.parse_file_as_json(f)

        # Collect counts by lane and sample (+source_files)
        self.bcl2fastq_bylane = dict()
        self.bcl2fastq_bysample = dict()
        self.bcl2fastq_bysample_lane = dict()
        self.source_files = dict()
        self.split_data_by_lane_and_sample()

        # Filter to strip out ignored sample names
        self.bcl2fastq_bylane = self.ignore_samples(self.bcl2fastq_bylane)
        self.bcl2fastq_bysample = self.ignore_samples(self.bcl2fastq_bysample)
        self.bcl2fastq_bysample_lane = self.ignore_samples(self.bcl2fastq_bysample_lane)

        # Return with Warning if no files are found
        if len(self.bcl2fastq_bylane) == 0 and len(self.bcl2fastq_bysample) == 0:
            raise ModuleNoSamplesFound

        # Print source files
        for s in self.source_files.keys():
            self.add_data_source(
                s_name=s,
                source=",".join(list(set(self.source_files[s]))),
                module="bcl2fastq",
                section="bcl2fastq-bysample",
            )

        # Add sample counts to general stats table
        self.add_general_stats()
        self.write_data_file(
            {str(k): self.bcl2fastq_bylane[k] for k in self.bcl2fastq_bylane.keys()}, "multiqc_bcl2fastq_bylane"
        )
        self.write_data_file(self.bcl2fastq_bysample, "multiqc_bcl2fastq_bysample")

        # Add section for summary stats per flow cell
        self.add_section(
            name="Lane Statistics",
            anchor="bcl2fastq-lanestats",
            description="Statistics about each lane for each flowcell",
            plot=self.lane_stats_table(),
        )

        # Add section for counts by lane
        cats = {
            "perfect": {"name": "Perfect Index Reads"},
            "imperfect": {"name": "Mismatched Index Reads"},
            "undetermined": {"name": "Undetermined Reads"},
        }
        self.add_section(
            name="Clusters by lane",
            anchor="bcl2fastq-bylane",
            description="Number of reads per lane (with number of perfect index reads).",
            helptext="""Perfect index reads are those that do not have a single mismatch.
                All samples of a lane are combined. Undetermined reads are treated as a third category.""",
            plot=bargraph.plot(
                self.get_bar_data_from_counts(self.bcl2fastq_bylane),
                cats,
                {
                    "id": "bcl2fastq_lane_counts",
                    "title": "bcl2fastq: Clusters by lane",
                    "ylab": "Number of clusters",
                    "hide_zero_cats": False,
                },
            ),
        )

        # Add section for counts by sample
        # get cats for per-lane tab
        lcats = set()
        for s_name in self.bcl2fastq_bysample_lane:
            lcats.update(self.bcl2fastq_bysample_lane[s_name].keys())
        lcats = sorted(list(lcats))
        self.add_section(
            name="Clusters by sample",
            anchor="bcl2fastq-bysample",
            description="Number of reads per sample.",
            helptext="""Perfect index reads are those that do not have a single mismatch.
                All samples are aggregated across lanes combined. Undetermined reads are ignored.
                Undetermined reads are treated as a separate sample.""",
            plot=bargraph.plot(
                [self.get_bar_data_from_counts(self.bcl2fastq_bysample), self.bcl2fastq_bysample_lane],
                [cats, lcats],
                {
                    "id": "bcl2fastq_sample_counts",
                    "title": "bcl2fastq: Clusters by sample",
                    "hide_zero_cats": False,
                    "ylab": "Number of clusters",
                    "data_labels": ["Index mismatches", "Counts per lane"],
                },
            ),
        )

        # Add section with undetermined barcodes
        self.add_section(
            name="Undetermined barcodes by lane",
            anchor="undetermined_by_lane",
            description="Count of the top twenty most abundant undetermined barcodes by lanes",
            plot=bargraph.plot(
                self.get_bar_data_from_undetermined(self.bcl2fastq_bylane),
                None,
                {
                    "id": "bcl2fastq_undetermined",
                    "title": "bcl2fastq: Undetermined barcodes by lane",
                    "ylab": "Reads",
                    "tt_percentages": False,
                    "use_legend": True,
                    "sort_samples": False,  # keep top barcode on top
                },
            ),
        )

    def parse_file_as_json(self, f):
        try:
            content = json.loads(f["f"])
        except ValueError:
            log.warning(f"Could not parse file as json: {f['fn']}")
            return

        # Clean / prepend directories to sample names
        runId = self.clean_s_name(content["RunId"], f)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None, runId)

        if runId not in self.bcl2fastq_data:
            self.bcl2fastq_data[runId] = dict()
        run_data = self.bcl2fastq_data[runId]
        for conversionResult in content.get("ConversionResults", []):
            lane_num = conversionResult["LaneNumber"]
            lane = f"L{conversionResult['LaneNumber']}"
            if lane in run_data:
                log.debug(f"Duplicate runId/lane combination found! Overwriting: {self.prepend_runid(runId, lane)}")
            run_data[lane] = {
                "total": 0,
                "total_yield": 0,
                "perfectIndex": 0,
                "samples": dict(),
                "yieldQ30": 0,
                "qscore_sum": 0,
            }
            # simplify the population of dictionaries
            rlane = run_data[lane]

            # Add undetermined barcodes
            unknown_barcode = dict()
            for lane_data in content.get("UnknownBarcodes", list()):
                if lane_data["Lane"] == lane_num:
                    unknown_barcode = lane_data["Barcodes"]
                    break
            run_data[lane]["unknown_barcodes"] = unknown_barcode

            for demuxResult in conversionResult.get("DemuxResults", []):
                if demuxResult["SampleName"] == demuxResult["SampleId"]:
                    sample = demuxResult["SampleName"]
                else:
                    sample = f"{demuxResult['SampleId']}-{demuxResult['SampleName']}"
                sample = self.clean_s_name(sample, f)
                if sample in run_data[lane]["samples"]:
                    log.debug(
                        "Duplicate runId/lane/sample combination found! Overwriting: {}, {}".format(
                            self.prepend_runid(runId, lane), sample
                        )
                    )
                run_data[lane]["samples"][sample] = {
                    "total": 0,
                    "total_yield": 0,
                    "perfectIndex": 0,
                    "filename": os.path.join(f["root"], f["fn"]),
                    "yieldQ30": 0,
                    "qscore_sum": 0,
                }
                # simplify the population of dictionaries
                lsample = run_data[lane]["samples"][sample]
                for r in range(1, 5):
                    lsample[f"R{r}_yield"] = 0
                    lsample[f"R{r}_Q30"] = 0
                    lsample[f"R{r}_trimmed_bases"] = 0
                rlane["total"] += demuxResult["NumberReads"]
                rlane["total_yield"] += demuxResult["Yield"]
                lsample["total"] += demuxResult["NumberReads"]
                lsample["total_yield"] += demuxResult["Yield"]
                for indexMetric in demuxResult.get("IndexMetrics", []):
                    rlane["perfectIndex"] += indexMetric["MismatchCounts"]["0"]
                    lsample["perfectIndex"] += indexMetric["MismatchCounts"]["0"]
                for readMetric in demuxResult.get("ReadMetrics", []):
                    r = readMetric["ReadNumber"]
                    rlane["yieldQ30"] += readMetric["YieldQ30"]
                    rlane["qscore_sum"] += readMetric["QualityScoreSum"]
                    lsample["yieldQ30"] += readMetric["YieldQ30"]
                    lsample["qscore_sum"] += readMetric["QualityScoreSum"]
                    lsample[f"R{r}_yield"] += readMetric["Yield"]
                    lsample[f"R{r}_Q30"] += readMetric["YieldQ30"]
                    lsample[f"R{r}_trimmed_bases"] += readMetric["TrimmedBases"]
                # Remove unpopulated read keys
                for r in range(1, 5):
                    if not lsample[f"R{r}_yield"] and not lsample[f"R{r}_Q30"] and not lsample[f"R{r}_trimmed_bases"]:
                        lsample.pop(f"R{r}_yield")
                        lsample.pop(f"R{r}_Q30")
                        lsample.pop(f"R{r}_trimmed_bases")
            undeterminedYieldQ30 = 0
            undeterminedQscoreSum = 0
            undeterminedTrimmedBases = 0
            if "Undetermined" in conversionResult:
                for readMetric in conversionResult["Undetermined"]["ReadMetrics"]:
                    undeterminedYieldQ30 += readMetric["YieldQ30"]
                    undeterminedQscoreSum += readMetric["QualityScoreSum"]
                    undeterminedTrimmedBases += readMetric["TrimmedBases"]
                run_data[lane]["samples"]["undetermined"] = {
                    "total": conversionResult["Undetermined"]["NumberReads"],
                    "total_yield": conversionResult["Undetermined"]["Yield"],
                    "perfectIndex": 0,
                    "yieldQ30": undeterminedYieldQ30,
                    "qscore_sum": undeterminedQscoreSum,
                    "trimmed_bases": undeterminedTrimmedBases,
                }

        # Calculate Percents and averages
        for lane_id, lane in run_data.items():
            try:
                lane["percent_Q30"] = (float(lane["yieldQ30"]) / float(lane["total_yield"])) * 100.0
            except ZeroDivisionError:
                lane["percent_Q30"] = "NA"
            try:
                lane["percent_perfectIndex"] = (float(lane["perfectIndex"]) / float(lane["total"])) * 100.0
            except ZeroDivisionError:
                lane["percent_perfectIndex"] = "NA"
            try:
                lane["mean_qscore"] = float(lane["qscore_sum"]) / float(lane["total_yield"])
            except ZeroDivisionError:
                lane["mean_qscore"] = "NA"

            for sample_id, sample in lane["samples"].items():
                try:
                    sample["percent_Q30"] = (float(sample["yieldQ30"]) / float(sample["total_yield"])) * 100.0
                except ZeroDivisionError:
                    sample["percent_Q30"] = "NA"
                try:
                    sample["percent_perfectIndex"] = (float(sample["perfectIndex"]) / float(sample["total"])) * 100.0
                except ZeroDivisionError:
                    sample["percent_perfectIndex"] = "NA"
                try:
                    sample["mean_qscore"] = float(sample["qscore_sum"]) / float(sample["total_yield"])
                except ZeroDivisionError:
                    sample["mean_qscore"] = "NA"

    def split_data_by_lane_and_sample(self):
        for run_id, r in self.bcl2fastq_data.items():
            for lane_id, lane in r.items():
                uniqLaneName = self.prepend_runid(run_id, lane_id)
                self.bcl2fastq_bylane[uniqLaneName] = {
                    "total": lane["total"],
                    "total_yield": lane["total_yield"],
                    "perfectIndex": lane["perfectIndex"],
                    "undetermined": lane["samples"].get("undetermined", {}).get("total", "NA"),
                    "yieldQ30": lane["yieldQ30"],
                    "qscore_sum": lane["qscore_sum"],
                    "percent_Q30": lane["percent_Q30"],
                    "percent_perfectIndex": lane["percent_perfectIndex"],
                    "mean_qscore": lane["mean_qscore"],
                    "unknown_barcodes": lane["unknown_barcodes"],
                }
                for sample_id, sample in lane["samples"].items():
                    if sample_id not in self.bcl2fastq_bysample:
                        self.bcl2fastq_bysample[sample_id] = {
                            "total": 0,
                            "total_yield": 0,
                            "perfectIndex": 0,
                            "yieldQ30": 0,
                            "qscore_sum": 0,
                        }
                        for r in range(1, 5):
                            self.bcl2fastq_bysample[sample_id][f"R{r}_yield"] = 0
                            self.bcl2fastq_bysample[sample_id][f"R{r}_Q30"] = 0
                            self.bcl2fastq_bysample[sample_id][f"R{r}_trimmed_bases"] = 0
                    s = self.bcl2fastq_bysample[sample_id]
                    s["total"] += sample["total"]
                    s["total_yield"] += sample["total_yield"]
                    s["perfectIndex"] += sample["perfectIndex"]
                    s["yieldQ30"] += sample["yieldQ30"]
                    s["qscore_sum"] += sample["qscore_sum"]
                    # Undetermined samples did not have R1 and R2 information
                    for r in range(1, 5):
                        try:
                            s[f"R{r}_yield"] += sample[f"R{r}_yield"]
                            s[f"R{r}_Q30"] += sample[f"R{r}_Q30"]
                            s[f"R{r}_trimmed_bases"] += sample[f"R{r}_trimmed_bases"]
                        except KeyError:
                            pass
                    try:
                        s["percent_Q30"] = (float(s["yieldQ30"]) / float(s["total_yield"])) * 100.0
                    except ZeroDivisionError:
                        s["percent_Q30"] = "NA"
                    try:
                        s["percent_perfectIndex"] = (float(s["perfectIndex"]) / float(s["total"])) * 100.0
                    except ZeroDivisionError:
                        s["percent_perfectIndex"] = "NA"
                    try:
                        s["mean_qscore"] = float(s["qscore_sum"]) / float(s["total_yield"])
                    except ZeroDivisionError:
                        s["mean_qscore"] = "NA"
                    if sample_id != "undetermined":
                        if sample_id not in self.source_files:
                            self.source_files[sample_id] = []
                        self.source_files[sample_id].append(sample["filename"])
                # Remove unpopulated read keys
                for sample_id, sample in lane["samples"].items():
                    for r in range(1, 5):
                        try:
                            if (
                                not self.bcl2fastq_bysample[sample_id][f"R{r}_yield"]
                                and not self.bcl2fastq_bysample[sample_id][f"R{r}_Q30"]
                                and not self.bcl2fastq_bysample[sample_id][f"R{r}_trimmed_bases"]
                            ):
                                self.bcl2fastq_bysample[sample_id].pop(f"R{r}_yield")
                                self.bcl2fastq_bysample[sample_id].pop(f"R{r}_Q30")
                                self.bcl2fastq_bysample[sample_id].pop(f"R{r}_trimmed_bases")
                        except KeyError:
                            pass

    def add_general_stats(self):
        data = dict()
        for sample_id, sample in self.bcl2fastq_bysample.items():
            percent_R_Q30 = dict()
            for r in range(1, 5):
                # Zero division is possible
                try:
                    percent_R_Q30[r] = "{0:.1f}".format(
                        float(100.0 * sample["R{}_Q30".format(r)] / sample["R{}_yield".format(r)])
                    )
                except ZeroDivisionError:
                    percent_R_Q30[r] = "0.0"
                except KeyError:
                    pass
            try:
                perfect_percent = f"{float(100.0 * sample['perfectIndex'] / sample['total']):.1f}"
            except ZeroDivisionError:
                perfect_percent = "0.0"

            data[sample_id] = {
                "yieldQ30": sample["yieldQ30"],
                "total": sample["total"],
                "perfectPercent": perfect_percent,
            }
            for r in range(1, 5):
                try:
                    data[sample_id][f"percent_R{r}_Q30"] = percent_R_Q30[r]
                    data[sample_id][f"R{r}_trimmed_bases"] = sample[f"R{r}_trimmed_bases"]
                except KeyError:
                    pass

        headers = {
            "total": {
                "title": f"{config.read_count_prefix} Clusters",
                "description": "Total number of reads for this sample as determined by bcl2fastq demultiplexing ({})".format(
                    config.read_count_desc
                ),
                "scale": "Blues",
                "shared_key": "read_count",
            },
            "yieldQ30": {
                "title": f"Yield ({config.base_count_prefix}) ≥ Q30",
                "description": f"Number of bases with a Phred score of 30 or higher ({config.base_count_desc})",
                "scale": "Greens",
                "shared_key": "base_count",
            },
        }
        # If no data for a column, header will be automatically ignored
        for r in range(1, 5):
            headers[f"percent_R{r}_Q30"] = {
                "title": f"% R{r} Yield ≥ Q30",
                "description": f"Percent of bases in R{r} with a Phred score of 30 or higher",
                "scale": "RdYlGn",
                "max": 100,
                "min": 0,
                "suffix": "%",
            }
        headers["perfectPercent"] = {
            "title": "% Perfect Index",
            "description": "Percent of reads with perfect index (0 mismatches)",
            "max": 100,
            "min": 0,
            "scale": "RdYlGn",
            "suffix": "%",
        }
        # If no data for a column, header will be automatically ignored
        for r in range(1, 5):
            hideCol = True
            for s in data:
                try:
                    if data[s][f"R{r}_trimmed_bases"] > 0:
                        hideCol = False
                except KeyError:
                    pass
            try:
                headers[f"R{r}_trimmed_bases"] = {
                    "title": f"{config.base_count_prefix} R{r} trimmed",
                    "description": f"Number of bases trimmed ({config.base_count_desc})",
                    "scale": "RdYlBu",
                    "modify": lambda x: x * 0.000001,
                    "hidden": hideCol,
                }
            except KeyError:
                pass
        self.general_stats_addcols(data, headers)

    def lane_stats_table(self):
        """Return a table with overview stats for each bcl2fastq lane for a single flow cell"""
        headers = {
            "total_yield": {
                "title": f"{config.base_count_prefix} Total Yield",
                "description": f"Number of bases ({config.base_count_desc})",
                "scale": "Greens",
                "shared_key": "base_count",
            },
            "total": {
                "title": f"{config.read_count_prefix} Total Clusters",
                "description": f"Total number of clusters for this lane ({config.read_count_desc})",
                "scale": "Blues",
                "shared_key": "read_count",
            },
            "percent_Q30": {
                "title": "% bases ≥ Q30",
                "description": "Percentage of bases with greater than or equal to Q30 quality score",
                "suffix": "%",
                "max": 100,
                "min": 0,
                "scale": "RdYlGn",
            },
            "mean_qscore": {
                "title": "Mean Quality",
                "description": "Average phred qualty score",
                "min": 0,
                "scale": "Spectral",
            },
            "percent_perfectIndex": {
                "title": "% Perfect Index",
                "description": "Percent of reads with perfect index (0 mismatches)",
                "max": 100,
                "min": 0,
                "scale": "RdYlGn",
                "suffix": "%",
            },
        }
        table_config = {
            "namespace": "bcl2fastq",
            "id": "bcl2fastq-lane-stats-table",
            "table_title": "bcl2fastq Lane Statistics",
            "col1_header": "Run ID - Lane",
        }
        return table.plot(self.bcl2fastq_bylane, headers, table_config)

    @staticmethod
    def prepend_runid(runId, rest):
        return str(runId) + " - " + str(rest)

    @staticmethod
    def get_bar_data_from_counts(data_by_flowcell):
        bar_data = {}
        for flowcell_id, data in data_by_flowcell.items():
            bar_data[flowcell_id] = {
                "perfect": data["perfectIndex"],
                "imperfect": data["total"] - data["perfectIndex"],
            }
            if "undetermined" in data:
                bar_data[flowcell_id]["undetermined"] = data["undetermined"]
        return bar_data

    @staticmethod
    def get_bar_data_from_undetermined(data_by_flowcell):
        """Get data to plot for undetermined barcodes."""
        bar_data = defaultdict(dict)
        # get undetermined barcodes for each lanes
        for flowcell_id, data in data_by_flowcell.items():
            try:
                for barcode, count in islice(data["unknown_barcodes"].items(), 20):
                    bar_data[barcode][flowcell_id] = count
            except AttributeError:
                pass

        sorted_items = sorted(bar_data.items(), key=lambda kv: sum(kv[1].values()), reverse=True)
        if len(sorted_items) > 20:
            sorted_items = sorted_items[:20]
        return dict(sorted_items)



""" MultiQC submodule to parse output from Picard TargetedPcrMetrics """

import logging

from multiqc.modules.picard import util
from multiqc.plots import bargraph
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard TargetedPcrMetrics reports and parse their data"""

    data_by_sample = dict()
    histogram_by_sample = dict()

    picard_config = getattr(config, "picard_config", {})
    skip_histo = picard_config.get("targeted_pcr_skip_histogram", False)

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/pcr_metrics", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]
        in_hist = False

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="TargetedPcrMetrics",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            if s_name is None:
                continue

            # Catch the histogram values
            if in_hist and not skip_histo:
                try:
                    sections = line.split("\t")
                    cov = int(sections[0])
                    count = int(sections[1])
                    histogram_by_sample[s_name][cov] = count
                except ValueError:
                    # Reset in case we have more in this log file
                    s_name = None
                    in_hist = False

            if util.is_line_right_before_table(line, picard_class="TargetedPcrMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")
                vals = f["f"].readline().strip("\n").split("\t")
                if len(vals) != len(keys):
                    continue

                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")

                module.add_data_source(f, s_name, section="TargetedPcrMetrics")
                data_by_sample[s_name] = dict()
                histogram_by_sample[s_name] = dict()

                for k, v in zip(keys, vals):
                    try:
                        # Multiply percentages by 100
                        if k.startswith("PCT_"):
                            v = float(v) * 100.0
                    except ValueError:
                        pass
                    data_by_sample[s_name][k] = v

            elif line.startswith("## HISTOGRAM"):
                keys = f["f"].readline().strip("\n").split("\t")
                assert len(keys) >= 2, (keys, f)
                in_hist = True
                histogram_by_sample[s_name] = dict()

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    histogram_by_sample = module.ignore_samples(histogram_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, "multiqc_picard_pcrmetrics")

    # Add to general stats table
    headers = {
        "PCT_AMPLIFIED_BASES": {
            "title": "Amplified Bases",
            "description": "The fraction of aligned bases that mapped to or near an amplicon.",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "scale": "BrBG",
        },
        "MEDIAN_TARGET_COVERAGE": {
            "title": "Median Target Coverage",
            "description": "The median coverage of reads that mapped to target regions of an experiment.",
            "min": 0,
            "suffix": "X",
            "scale": "GnBu",
        },
    }
    module.general_stats_addcols(data_by_sample, headers, namespace="TargetedPcrMetrics")

    # Bar plot of ignored bases
    keys = {
        "ON_AMPLICON_BASES": {"name": "On-amplicon bases"},
        "NEAR_AMPLICON_BASES": {"name": "Near-amplicon bases"},
        "OFF_AMPLICON_BASES": {"name": "Off-amplicon bases", "color": "#f28f43"},
    }

    # Config for the plot
    pconfig = {
        "id": "picard_pcr_metrics_bases",
        "title": "Picard: PCR Amplicon Bases",
        "ylab": "# Bases",
        "cpswitch_counts_label": "# Bases",
        "hide_zero_cats": False,
    }

    module.add_section(
        name="PCR Amplicon Bases",
        anchor="picard-pcrmetrics-bases",
        description="Metrics about reads obtained from targeted PCR experiments.",
        helptext="""
        This plot shows the number of bases aligned on or near to amplified regions of the genome.

        * `ON_AMPLICON_BASES`: The number of `PF_BASES_ALIGNED` that mapped to an amplified region of the genome.
        * `NEAR_AMPLICON_BASES`: The number of `PF_BASES_ALIGNED` that mapped to within a fixed interval of an amplified region, but not on a baited region.
        * `OFF_AMPLICON_BASES`: The number of `PF_BASES_ALIGNED` that mapped neither on or near an amplicon.

        For more information see the [Picard documentation](https://broadinstitute.github.io/picard/picard-metric-definitions.html#TargetedPcrMetrics).""",
        plot=bargraph.plot(data_by_sample, keys, pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC module to parse output from BioBloom Tools """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="BioBloom Tools",
            anchor="biobloomtools",
            href="https://github.com/bcgsc/biobloom/",
            info="creates filters for a genome reference and categorises "
            "sequences. This is faster than alignment and can be used for pre-processing "
            "and QC applications such as contamination detection.",
            doi="10.1093/bioinformatics/btu558",
        )

        # Find and load any BioBloom Tools reports
        self.bbt_data = dict()
        self.num_orgs = 0
        for f in self.find_log_files("biobloomtools", filehandles=True):
            parsed_data = self.parse_bbt(f["f"])
            if len(parsed_data) > 0:
                if f["s_name"] in self.bbt_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f)
                self.bbt_data[f["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.bbt_data = self.ignore_samples(self.bbt_data)

        if len(self.bbt_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.bbt_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Section 1 - Alignment Profiles
        self.add_section(plot=self.bbt_simple_plot())

        # Write the total counts and percentages to files
        data_export = {}
        for s_name in self.bbt_data:
            data_export[s_name] = dict()
            for org in self.bbt_data[s_name]:
                data_export[s_name][org] = int(self.bbt_data[s_name][org]["hits"])
        self.write_data_file(data_export, "multiqc_biobloomtools")

    def parse_bbt(self, fh):
        """Parse the BioBloom Tools output into a 3D dict"""
        parsed_data = dict()
        headers = None
        for line in fh:
            s = line.split("\t")
            if headers is None:
                headers = s
            else:
                parsed_data[s[0]] = dict()
                for i, h in enumerate(headers[1:]):
                    parsed_data[s[0]][h] = float(s[i + 1])

        return parsed_data

    def bbt_simple_plot(self):
        """Makes a simple bar plot with summed alignment counts for
        each species, stacked."""

        # First, sum the different types of alignment counts
        data = dict()
        cats = dict()
        for s_name in self.bbt_data:
            data[s_name] = dict()
            for org in self.bbt_data[s_name]:
                data[s_name][org] = self.bbt_data[s_name][org]["hits"] - self.bbt_data[s_name][org]["shared"]
                if org not in cats and org != "multiMatch" and org != "noMatch":
                    if org.lower().endswith(".fa"):
                        cname = org[:-3]
                    elif org.lower().endswith(".fasta"):
                        cname = org[:-6]
                    else:
                        cname = org
                    cats[org] = {"name": cname}

        pconfig = {
            "id": "biobloom_tools",
            "title": "BioBloom Tools: Alignment counts per species",
            "ylab": "Number of hits",
            "hide_zero_cats": False,
        }
        cats["multiMatch"] = {"name": "Multiple Genomes", "color": "#820000"}
        cats["noMatch"] = {"name": "No Match", "color": "#cccccc"}

        return bargraph.plot(data, cats, pconfig)



""" MultiQC module to parse log output from Xenome Classify """

from collections import defaultdict

import logging
import spectra
from typing import Dict, Union

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Xenome",
            anchor="xenome",
            href="https://github.com/data61/gossamer/blob/master/docs/xenome.md",
            info="is a tool for classifying reads from xenograft sources.",
            doi="doi.org/10.1093/bioinformatics/bts236",
        )

        # Find and load any Xenome reports
        self.detail_percents = dict()
        self.detail_counts = dict()
        self.summary_percents = dict()
        self.summary_counts = dict()
        for f in self.find_log_files("xenome"):
            self._parse_xenome_logs(f)

        # Filter to strip out ignored sample names
        self.detail_percents = self.ignore_samples(self.detail_percents)
        self.detail_counts = self.ignore_samples(self.detail_counts)
        self.summary_percents = self.ignore_samples(self.summary_percents)
        self.summary_counts = self.ignore_samples(self.summary_counts)
        if len(self.detail_percents) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.detail_percents)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.detail_percents, "multiqc_xenome_detail_percent")
        self.write_data_file(self.detail_counts, "multiqc_xenome_detail_counts")
        self.write_data_file(self.summary_percents, "multiqc_xenome_summary_percent")
        self.write_data_file(self.summary_counts, "multiqc_xenome_summary_counts")

        self.all_species = self._collect_all_species(self.summary_counts)
        self._build_table()
        self._xenome_stats_plot()

    @staticmethod
    def _collect_all_species(summary_data):
        grafts = []
        hosts = []
        for sn, cnt_by_cls in summary_data.items():
            s = list(cnt_by_cls)[0]
            if s not in grafts:
                grafts.append(s)
            s = list(cnt_by_cls)[1]
            if s not in hosts:
                hosts.append(s)
        return grafts + [s for s in hosts if s not in grafts]

    def _parse_xenome_logs(self, f):
        """
        Expected two sections: "Statistics" and "Summary", one following the other after a line break
        """
        s_name = f["s_name"]

        lines = iter(f["contents_lines"])
        try:
            detail_percents, detail_counts = self._parse_xenome_section(lines, "Statistics")
            summary_percents, summary_counts = self._parse_xenome_section(lines, "Summary")
        except (AssertionError, StopIteration) as e:
            log.error(f"Error parsing Xenome log file '{f['fn']}' for sample '{s_name}': {e}")
            return

        if s_name in self.detail_counts:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.add_data_source(f, s_name)
        self.detail_counts[s_name] = detail_counts
        self.detail_percents[s_name] = detail_percents
        self.summary_counts[s_name] = summary_counts
        self.summary_percents[s_name] = summary_percents

    @staticmethod
    def _parse_xenome_section(lines, expected_title: str):
        """
        Parse one section, e.g.:

        Statistics
        B	G	H	M	count	percent	class
        0	0	0	0	73938	0.231509	"neither"
        ...

        OR

        Summary
        count	percent	class
        30220658	94.6246	human
        ...
        """
        title = next(lines).strip()
        assert title == expected_title, f"Expected '{expected_title}', got '{title}'"
        header = next(lines).strip().split("\t")
        expected_fields = ["count", "percent", "class"]
        assert all(
            field in header for field in expected_fields
        ), f"Expected the header to contain '{expected_fields}', got {header}"

        cnt_by_class = defaultdict(int)
        pct_by_class = defaultdict(float)
        for line in lines:
            if line.strip() == "":
                break
            fields = line.strip().split("\t")
            assert len(fields) == len(header), f"Expected {len(header)} fields, got {len(fields)}"
            data = dict(zip(header, fields))
            cls = data["class"].strip('"')
            cnt_by_class[cls] += int(data["count"])
            pct_by_class[cls] += float(data["percent"])
        return pct_by_class, cnt_by_class

    # to support multiple species, when running on logs for many unrelated runs
    COLORS = [
        ("#377eb8", "Blues"),  # blue
        ("#e41a1c", "Reds"),  # red
        ("#4daf4a", "Greens"),  # green
        ("#ff7f00", "Oranges"),  # orange
        ("#ffff33", "YlOrRd"),  # yellow
        ("#a65628", "YlOrBr"),  # brown
        ("#f781bf", "PuRd"),  # pink
    ]

    BOTH = ("#984ea3", "Purples")  # purple
    AMBIGUOUS = ("#616161", "Greys")  # grey
    NEITHER = ("#b3b3b3", "Greys")  # light grey

    @staticmethod
    def _lighten_color(code: str, lighten=1.0):
        def _rgb_converter(x):
            return max(0, min(1, 1 + ((x - 1) * lighten)))

        c = spectra.html(code)
        c = spectra.rgb(*[_rgb_converter(v) for v in c.rgb])
        return c.hexcode

    def _get_color(self, cls: Union[str, int], lighten=1.0, return_scale=False) -> str:
        if not isinstance(cls, int) and cls in self.all_species:
            cls = self.all_species.index(cls)
        if isinstance(cls, int):
            code, scale = MultiqcModule.COLORS[cls % len(MultiqcModule.COLORS)]
        elif cls == "both":
            code, scale = MultiqcModule.BOTH
        elif cls == "ambiguous":
            code, scale = MultiqcModule.AMBIGUOUS
        else:
            code, scale = MultiqcModule.NEITHER

        if return_scale:
            return scale
        else:
            return MultiqcModule._lighten_color(code, lighten)

    def _build_table(self):
        """
        Prepare headers and data for a table. Add a section with a table,
        and add a few columns into the general stats.
        """
        headers: Dict[str, Dict] = {}
        table_data = defaultdict(dict)

        for sn, data in self.summary_percents.items():
            for cls, val in data.items():
                table_data[sn][f"{cls}_reads_pct"] = val
                if cls == "human":
                    headers[f"{cls}_reads_pct"] = {
                        "rid": f"{self.anchor}_{cls}_reads_pct",  # to make the ID unique from xengsort
                        "title": "Human reads",
                        "description": "share of human reads in the sample",
                        "min": 0,
                        "suffix": "%",
                        "scale": self._get_color(cls, return_scale=True),
                        "format": "{:,.1f}",
                    }
                else:
                    headers[f"{cls}_reads_pct"] = {
                        "rid": f"{self.anchor}_{cls}_reads_pct",  # to make the ID unique from xengsort
                        "title": f"{cls.capitalize()} reads",
                        "description": f"share of {cls} reads in the sample",
                        "min": 0,
                        "suffix": "%",
                        "scale": self._get_color(cls, return_scale=True),
                        "format": "{:,.1f}",
                        "hidden": cls in ["both", "neither", "ambiguous"],
                    }
        self.general_stats_addcols(table_data, headers)
        detail_headers = headers.copy()
        for metric in headers:
            detail_headers[metric]["hidden"] = False

        for sn, data in self.summary_counts.items():
            for cls, val in data.items():
                table_data[sn][f"{cls}_reads_cnt"] = val
                if cls == "human":
                    detail_headers[f"{cls}_reads_cnt"] = {
                        "title": "Human reads",
                        "description": "number of human reads in the sample",
                        "min": 0,
                        "scale": self._get_color(cls, return_scale=True),
                        "format": "{:,d}",
                        "hidden": True,
                    }
                else:
                    detail_headers[f"{cls}_reads_cnt"] = {
                        "title": f"{cls.capitalize()} reads",
                        "description": f"number of {cls} reads in the sample",
                        "min": 0,
                        "scale": self._get_color(cls, return_scale=True),
                        "format": "{:,d}",
                        "hidden": True,
                    }

        self.add_section(
            name="Summary table",
            anchor="xenome-summary-table",
            plot=table.plot(
                table_data,
                detail_headers,
                pconfig={
                    "id": "xenome-table",
                    "table_title": "Xenome",
                },
            ),
        )

    def _xenome_stats_plot(self):
        """
        Create two bar plots: based on summary and detail data.
        """
        colors = {s: self._get_color(i, lighten=0.6) for i, s in enumerate(self.all_species)}
        definitely_colors = {s: self._get_color(i, lighten=1.0) for i, s in enumerate(self.all_species)}
        probably_colors = {s: self._get_color(i, lighten=0.3) for i, s in enumerate(self.all_species)}

        summary_cats = {}
        detail_cats = {}
        for s in self.all_species:
            summary_cats[s] = {"name": s.capitalize(), "color": colors[s]}
            detail_cats[f"definitely {s}"] = {"name": f"Definitely {s}", "color": definitely_colors[s]}
            detail_cats[f"probably {s}"] = {"name": f"Probably {s}", "color": probably_colors[s]}

        summary_cats["both"] = {"name": "Both", "color": self._get_color("both", lighten=0.6)}
        summary_cats["ambiguous"] = {"name": "Ambiguous", "color": self._get_color("ambiguous")}
        summary_cats["neither"] = {"name": "Neither", "color": self._get_color("neither")}
        detail_cats["both"] = summary_cats["both"]
        detail_cats["probably both"] = {
            "name": "Probably both",
            "color": self._get_color("both", 0.3),
        }
        detail_cats["ambiguous"] = summary_cats["ambiguous"]
        detail_cats["neither"] = summary_cats["neither"]

        self.add_section(
            description="This plot shows the number of reads classified by Xenome",
            helptext="""
            There are 5 possible categories:  
            * Reads found in graft species, e.g. **Human**
            * Reads found in host species, e.g. **Mouse**  
            * **Both**: read was found in either of the species
            * **Neither**: Read was found in neither of the species
            * **Ambiguous**: Read origin could not be adequately determined.  
            """,
            name="Summary classification",
            anchor="xenome_summary_bar_plot_section",
            plot=bargraph.plot(
                self.summary_counts,
                summary_cats,
                {
                    "id": "xenome_summary_bar_plot",
                    "title": "Xenome: summary classification",
                    "ylab": "# Reads",
                    "cpswitch_counts_label": "Number of reads",
                    "cpswitch_c_active": False,
                    "logswitch": True,
                },
            ),
        )

        self.add_section(
            description="This plot shows the number of reads classified by Xenome, "
            "with a more detail certainty split than the plot above",
            name="Detailed classification",
            anchor="xenome_detail_bar_plot_section",
            plot=bargraph.plot(
                self.detail_counts,
                detail_cats,
                {
                    "id": "xenome_detail_bar_plot",
                    "title": "Xenome: detailed classification",
                    "ylab": "# Reads",
                    "cpswitch_counts_label": "Number of reads",
                    "cpswitch_c_active": False,
                    "logswitch": True,
                },
            ),
        )



"""MultiQC module to parse CheckQC JSON output"""

import json
import logging
import re
from operator import itemgetter

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

log = logging.getLogger(__name__)

handlers = (
    "ClusterPFHandler",
    "Q30Handler",
    "ErrorRateHandler",
    "ReadsPerSampleHandler",
    "UnidentifiedIndexHandler",
    "UndeterminedPercentageHandler",
)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="CheckQC",
            anchor="checkqc",
            href="https://github.com/Molmed/checkQC",
            info="is a program designed to check a set of quality criteria against an Illumina runfolder. Samples are only shown in the report if they fail a check.",
            doi="10.21105/joss.00556",
        )

        self.checkqc_data = dict()
        self.runs = set()

        self.log_files = list(self.find_log_files("checkqc"))
        self.onerun = len(self.log_files) == 1

        for f in self.log_files:
            raw_content = json.loads(f["f"])
            self.parse_checkqc_json(raw_content, f)
            self.add_data_source(f)

        if not self.checkqc_data:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.log_files)} run and {len(self.checkqc_data)} samples")

        self.write_data_file(self.checkqc_data, "multiqc_checkqc")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.add_sections()

    def parse_checkqc_json(self, content, f):
        run = self._get_unique_runname(content)
        if "ReadsPerSampleHandler" in content:
            self.get_reads_per_sample_data(content["ReadsPerSampleHandler"], run, f)
        if "ClusterPFHandler" in content:
            self.get_cluster_pf_data(content["ClusterPFHandler"], run, f)
        if "Q30Handler" in content:
            self.get_q30_data(content["Q30Handler"], run, f)
        if "ErrorRateHandler" in content:
            self.get_error_rate_data(content["ErrorRateHandler"], run, f)
        if "UndeterminedPercentageHandler" in content:
            self.get_undetermined_percentage_data(content["UndeterminedPercentageHandler"], run, f)
        if "UnidentifiedIndexHandler" in content:
            self.get_unidentified_index_data(content["UnidentifiedIndexHandler"], run, f)

    def _get_unique_runname(self, content):
        base_run_name = content["run_summary"]["instrument_and_reagent_type"]
        run_name = base_run_name
        i = 1
        while run_name in self.runs:
            run_name = f"{base_run_name}_{i}"
            i += 1
        self.runs.add(run_name)
        return run_name

    def add_sections(self):
        if "ReadsPerSampleHandler" in self.checkqc_data:
            self.add_reads_per_sample_section()
        if "ClusterPFHandler" in self.checkqc_data:
            self.add_cluster_pf_section()
        if "Q30Handler" in self.checkqc_data:
            self.add_q30_section()
        if "ErrorRateHandler" in self.checkqc_data:
            self.add_error_rate_section()
        if "UndeterminedPercentageHandler" in self.checkqc_data:
            self.add_undetermined_percentage_section()
        if "UndeterminedPercentageHandler_ZeroYield" in self.checkqc_data:
            self.add_undetermined_percentage_zero_yield_section()
        if "UnidentifiedIndexHandler" in self.checkqc_data:
            self.add_unidentified_index_section()

    def get_reads_per_sample_data(self, issues, run, f):
        """Parse data from checkQC ReadsPerSampleHandler

        Args:
            issues (dict): JSON dict from CheckQC containing ReadsPerSampleHandler results
            run (str): name of sequencing run
            f (dict): MultiQC log file dict
        """
        data = {}
        for issue in issues:
            is_error = issue["type"] == "error"
            sample_name = issue["data"]["sample_name"]
            lane = issue["data"]["lane"]
            if self.onerun:
                sample = self.clean_s_name(f"{sample_name} - Lane {lane}", f)
            else:
                sample = self.clean_s_name(f"{sample_name} - Lane {lane}, run {run}", f)
            if self.is_ignore_sample(sample):
                continue
            self.add_data_source(f, sample)

            read_num = issue["data"]["sample_reads"]
            read_threshold = issue["data"]["threshold"]

            data[sample] = {"read_num": read_num, "threshold": read_threshold}
            if is_error:
                data[sample]["missing_error"] = read_threshold - read_num
            else:
                data[sample]["missing_warning"] = read_threshold - read_num
        if data:
            if "ReadsPerSampleHandler" not in self.checkqc_data:
                self.checkqc_data["ReadsPerSampleHandler"] = data
            else:
                self.checkqc_data["ReadsPerSampleHandler"].update(data)

    def add_reads_per_sample_section(self):
        """Add a section for samples with too few reads

        Creates a barplot with read number and number missing to reach threshold

        """
        data = self.checkqc_data["ReadsPerSampleHandler"]

        warning, error = self._get_warning_error(data)

        cats = dict()
        cats["read_num"] = {
            "name": "Reads",
        }
        if warning:
            cats["missing_warning"] = {"name": "Reads missing to reach threshold for QC warning", "color": "#ffc300"}
        if error:
            cats["missing_error"] = {"name": "Reads missing to reach threshold for QC error", "color": "#f44336"}

        pconfig = {
            "id": "checkqc_reads-per-sample-plot",
            "title": "CheckQC: Number reads too low",
            "ylab": "Number of reads [M]",
            "xlab": "Sample - Lane",
        }

        self.add_section(
            name="Too few reads per sample",
            anchor="checkqc-readspersample",
            description="Checks that the number of reads assigned to a sample is high enough. See [documentation](https://checkqc.readthedocs.io/en/latest/checkQC.handlers.reads_per_sample_handler.html).",
            helptext="""The value specified in the configuration is interpreted as the number of reads demanded for a single sample,
            i.e. the number of reads per sample on a lane which has multiple samples is the threshold divided by the total number of samples on the lane.
            """,
            plot=bargraph.plot(data, cats, pconfig),
        )

    def get_cluster_pf_data(self, issues, run, f):
        """Get data from ClusterPFHandler

        Args:
            issues (dict): JSON dict from CheckQC containing ClusterPFHandler results
            run (str): name of sequencing run
            f (dict): MultiQC log file dict
        """
        data = {}
        for issue in issues:
            is_error = issue["type"] == "error"
            lane = str(issue["data"]["lane"])
            if self.onerun:
                sample = self.clean_s_name(lane, f)
            else:
                sample = self.clean_s_name(f"{lane} ({run})", f)

            lane_pf = issue["data"]["lane_pf"]
            threshold = issue["data"]["threshold"] * pow(10, 6)
            if self.is_ignore_sample(sample):
                continue
            self.add_data_source(f, sample)

            data[sample] = {"lane_pf": lane_pf}
            if is_error:
                data[sample]["missing_error"] = threshold - lane_pf
            else:
                data[sample]["missing_warning"] = threshold - lane_pf
        if data:
            if "ClusterPFHandler" not in self.checkqc_data:
                self.checkqc_data["ClusterPFHandler"] = data
            else:
                self.checkqc_data["ClusterPFHandler"].update(data)

    def add_cluster_pf_section(self):
        """Add a section for lanes with cluster pf too low

        Creates a barplot with cluster pf number and number missing to reach threshold

        """
        data = self.checkqc_data["ClusterPFHandler"]
        data = {f"Lane {k}": v for k, v in data.items()}

        warning, error = self._get_warning_error(data)

        cats = dict()
        cats["lane_pf"] = {
            "name": "Clusters passing filters",
        }

        if error:
            cats["missing_error"] = {"name": "Cluster PF missing to reach threshold for QC error", "color": "#f44336"}
        if warning:
            cats["missing_warning"] = {
                "name": "Cluster PF missing to reach threshold for QC warning",
                "color": "#ffc300",
            }

        pconfig = {
            "id": "checkqc_cluster-pf-plot",
            "title": "CheckQC: Too few clusters passing filters",
            "ylab": "Number of clusters",
        }

        self.add_section(
            name="Clusters passing filters",
            anchor="checkqc-clusterpf",
            description="Checks that the number of clusters passing filter (PF) on a sequencing lane passes the set criteria. See [docuementation](https://checkqc.readthedocs.io/en/latest/checkQC.handlers.cluster_pf_handler.html).",
            plot=bargraph.plot(data, cats, pconfig),
        )

    def get_q30_data(self, issues, run, f):
        """Get data from Q30Handler

        Args:
            issues (dict): JSON dict from CheckQC containing Q30Handler results
            run (str): name of sequencing run
            f (dict): MultiQC log file dict
        """
        data = {}
        for issue in issues:
            is_error = issue["type"] == "error"
            lane = issue["data"]["lane"]
            read = issue["data"]["read"]
            if self.onerun:
                sample = self.clean_s_name(f"{lane} - R{read}", f)
            else:
                sample = self.clean_s_name(f"{lane} - R{read} ({run})", f)

            if self.is_ignore_sample(sample):
                continue
            self.add_data_source(f, sample)
            percent_q30 = issue["data"]["percent_q30"]
            threshold = issue["data"]["threshold"]
            data[sample] = {"percent_q30": percent_q30}
            if is_error:
                data[sample]["missing_error"] = threshold - percent_q30
            else:
                data[sample]["missing_warning"] = threshold - percent_q30
        if data:
            if "Q30Handler" not in self.checkqc_data:
                self.checkqc_data["Q30Handler"] = data
            else:
                self.checkqc_data["Q30Handler"].update(data)

    def add_q30_section(self):
        """Add a section for lanes with Q30 too low

        Creates a barplot with %Q30 and percentage missing to reach threshold

        """
        data = self.checkqc_data["Q30Handler"]
        data = {f"Lane {k}": v for k, v in data.items()}

        warning, error = self._get_warning_error(data)

        cats = dict()
        cats["percent_q30"] = {
            "name": "%Q30",
        }
        if warning:
            cats["missing_warning"] = {"name": "%Q30 missing to reach threshold for QC warning", "color": "#ffc300"}
        if error:
            cats["missing_error"] = {"name": "%Q30 missing to reach threshold for QC error", "color": "#f44336"}

        pconfig = {
            "id": "checkqc_q30-plot",
            "title": "CheckQC: %Q30 too low",
            "ylab": "%Q30",
            "xlab": "Lane - Read",
            "cpswitch": False,
        }

        self.add_section(
            name="%Q30 too low",
            anchor="checkqc-q30",
            description="Checks that the percentage of bases in read 1 or read 2 with base quality over 30 (%Q30) passes the set criteria. See [documentation](https://checkqc.readthedocs.io/en/latest/checkQC.handlers.q30_handler.html).",
            plot=bargraph.plot(data, cats, pconfig),
        )

    def get_error_rate_data(self, issues, run, f):
        """Get data from ErrorRateHandler

        Args:
            issues (dict): JSON dict from CheckQC containing ErrorRateHandler results
            run (str): name of sequencing run
            f (dict): MultiQC log file dict
        """
        data = {}
        for issue in issues:
            is_error = issue["type"] == "error"
            lane = issue["data"]["lane"]
            read = issue["data"]["read"]
            if self.onerun:
                sample = self.clean_s_name(f"{lane} - R{read}", f)
            else:
                sample = self.clean_s_name(f"{lane} - R{read} ({run})", f)

            if self.is_ignore_sample(sample):
                continue
            self.add_data_source(f, sample)
            error_rate = issue["data"]["error_rate"]
            threshold = issue["data"]["threshold"]
            data[sample] = {"threshold": threshold}
            if is_error:
                data[sample]["missing_error"] = error_rate - threshold
            else:
                data[sample]["missing_warning"] = error_rate - threshold
        if data:
            if "ErrorRateHandler" not in self.checkqc_data:
                self.checkqc_data["ErrorRateHandler"] = data
            else:
                self.checkqc_data["ErrorRateHandler"].update(data)

    def add_error_rate_section(self):
        """Add a section for lanes with error rate too high

        Creates a barplot with error rate and missing to reach threshold

        """
        data = self.checkqc_data["ErrorRateHandler"]
        data = {f"Lane {k}": v for k, v in data.items()}

        warning, error = self._get_warning_error(data)

        cats = dict()
        cats["threshold"] = {
            "name": "Error rate part until threshold",
        }
        if warning:
            cats["missing_warning"] = {"name": "Error rate part above threshold for QC warning", "color": "#ffc300"}
        if error:
            cats["missing_error"] = {"name": "Error rate part above threshold for QC error", "color": "#f44336"}

        pconfig = {
            "id": "checkqc_error-rate-plot",
            "title": "CheckQC: Error rate too high",
            "ylab": "Error rate",
            "xlab": "Lane - Read",
            "cpswitch": False,
        }

        self.add_section(
            name="Error rate too high",
            anchor="checkqc-errorrate",
            description="Checks that the sequencing error rate per lane and read are below the specified threshold. See [documentation](https://checkqc.readthedocs.io/en/latest/checkQC.handlers.error_rate_handler.html).",
            plot=bargraph.plot(data, cats, pconfig),
        )

    def get_undetermined_percentage_data(self, issues, run, f):
        """Get data from UndeterminedPercentageHandler

        Args:
            issues (dict): JSON dict from CheckQC containing UndeterminedPercentageHandler results
            run (str): name of sequencing run
            f (dict): MultiQC log file dict
        """
        zero_yield_data = {}
        data = {}
        for issue in issues:
            is_error = issue["type"] == "error"
            lane = str(issue["data"]["lane"])
            if self.onerun:
                sample = self.clean_s_name(lane, f)
            else:
                sample = self.clean_s_name(f"{lane} ({run})", f)

            if self.is_ignore_sample(sample):
                continue
            self.add_data_source(f, sample)
            p_undetermined = issue["data"]["percentage_undetermined"]
            if p_undetermined == "N/A":
                zero_yield_data[run] = {"lane": lane}
            else:
                threshold = issue["data"]["threshold"]
                computed_threshold = issue["data"]["computed_threshold"]
                phix = issue["data"]["phix_on_lane"]
                data[sample] = {"phix": phix, "threshold": threshold}
                if is_error:
                    data[sample]["missing_error"] = p_undetermined - computed_threshold
                else:
                    data[sample]["missing_warning"] = p_undetermined - computed_threshold
        if data:
            if "UndeterminedPercentageHandler" not in self.checkqc_data:
                self.checkqc_data["UndeterminedPercentageHandler"] = data
            else:
                self.checkqc_data["UndeterminedPercentageHandler"].update(data)
        if zero_yield_data:
            if "UndeterminedPercentageHandler_ZeroYield" not in self.checkqc_data:
                self.checkqc_data["UndeterminedPercentageHandler_ZeroYield"] = zero_yield_data
            else:
                self.checkqc_data["UndeterminedPercentageHandler_ZeroYield"].update(zero_yield_data)

    def add_undetermined_percentage_section(self):
        """Add a section for lanes with undetermined index percentage too high

        Creates a barplot with undetermined index percentage and missing to reach threshold

        Args:
            run (str): name of sequencing run
        """
        data = self.checkqc_data["UndeterminedPercentageHandler"]
        data = {f"Lane {k}": v for k, v in data.items()}

        warning, error = self._get_warning_error(data)
        cats = dict()
        cats["phix"] = {"name": r"% PhiX", "color": "#88a680"}
        cats["threshold"] = {
            "name": r"% undetermined indexes until threshold",
        }
        if warning:
            cats["missing_warning"] = {
                "name": r"% undetermined indexes above threshold for QC warning",
                "color": "#ffc300",
            }
        if error:
            cats["missing_error"] = {"name": r"% undetermined indexes above threshold for QC error", "color": "#f44336"}

        pconfig = {
            "id": "checkqc_undetermined-percentage-plot",
            "title": "CheckQC: Percentage undetermined indexes too high",
            "ylab": r"% undetermined indexes",
            "cpswitch": False,
        }

        self.add_section(
            name="% Undetermined indexes",
            anchor="checkqc-undetermined-rate",
            description="This checks that the percentage of undetermined reads on a lane is below the specified threshold. If there are no indexes specified for the lane, this will be skipped. See [documentation](https://checkqc.readthedocs.io/en/latest/checkQC.handlers.undetermined_percentage_handler.html).",
            plot=bargraph.plot(data, cats, pconfig),
        )

    def add_undetermined_percentage_zero_yield_section(self):
        """Add a section for lanes with zero yield

        Creates a table with the lane names with zero yield

        Args:
            run (str): name of sequencing run
        """
        data = self.checkqc_data["UndeterminedPercentageHandler_ZeroYield"]

        pconfig = {
            "id": "checkqc_zero-yield-table",
            "table_title": "CheckQC: Lanes with Yield 0",
            "scale": "Reds",
            "col1_header": "Run",
        }
        headers = {
            "lane": {
                "title": "Lane",
                "description": "Sequencing lane",
                "format": "{:,.0f}",
            },
        }

        self.add_section(
            name="Lanes with zero yield",
            anchor="checkqc-zero-yield",
            description="Some lanes have zero yield, no undetermined percentage computation possible.",
            plot=table.plot(data, headers, pconfig),
        )

    def get_unidentified_index_data(self, issues, run, f):
        """Get data for UnidentifiedIndexHandler

        Args:
            issues (dict): JSON dict from CheckQC containing UnidentifiedIndexHandler results
            run (str): name of sequencing run
            f (dict): MultiQC log file dict
        """
        idx_to_lane_to_rep = {}

        for issue in issues:
            msg = issue["data"]["msg"]
            m = re.match(
                r"Index: ([ATGC+]+) on lane: (\d+) was significantly "
                r"overrepresented \(([0-9.]+)%\) at significance "
                r"threshold of: [0-9.]+%\.",
                msg,
            )
            if not m:
                # TODO: Could add handling of other message types
                continue
            index, lane, overrep = m.groups()
            overrep = float(overrep)

            if index not in idx_to_lane_to_rep:
                idx_to_lane_to_rep[index] = {}
            idx_to_lane_to_rep[index][lane] = overrep

        # Sort indexes by average overrepresentation and take top 20
        sorted_idx = sorted(
            [(idx, sum(l2r.values()) / len(l2r)) for idx, l2r in idx_to_lane_to_rep.items()],
            key=itemgetter(1),
            reverse=True,
        )[:20]

        data = {}
        for idx, _ in sorted_idx:
            sample = self.clean_s_name(idx, f)

            if self.is_ignore_sample(sample):
                continue
            self.add_data_source(f, sample)
            if sample not in data:
                data[sample] = {}
            for lane, val in idx_to_lane_to_rep[idx].items():
                if self.onerun:
                    data[sample][lane] = val
                else:
                    data[sample][f"{lane} ({run})"] = val

        if data:
            if "UnidentifiedIndexHandler" not in self.checkqc_data:
                self.checkqc_data["UnidentifiedIndexHandler"] = data
            else:
                self.checkqc_data["UnidentifiedIndexHandler"].update(data)

    def add_unidentified_index_section(self):
        """Add a section for lanes with unidentified index percentage too high

        Creates a barplot with undetermined index overrepresentation by lane

        Args:
            run (str): name of sequencing run
        """
        data = self.checkqc_data["UnidentifiedIndexHandler"]

        lanes = set()
        for sample in data:
            for lane in data[sample]:
                lanes.add(lane)

        cats = dict()
        for lane in sorted(lanes):
            cats[lane] = {"name": f"Lane {lane}"}

        pconfig = {
            "id": "checkqc_unidentified-index-plot",
            "title": "CheckQC: Overrepresented unidentified indexes",
            "ylab": "% representation",
            "xlab": "Overrepresented indexes",
            "stacking": None,
        }

        self.add_section(
            name="Unidentified indexes",
            anchor="checkqc-unidentifiedpercentage",
            description="This check attempts to identify any indexes that are highly represented within unidentified reads. See the [documentation](https://checkqc.readthedocs.io/en/latest/checkQC.handlers.unidentified_index_handler.html).",
            helptext="This check ignores any indexes which have `N`’s in them. These are assumed to be read errors.",
            plot=bargraph.plot(data, cats, pconfig),
        )

    @staticmethod
    def _get_warning_error(data):
        warning = False
        error = False
        for sample in data:
            if "missing_error" in data[sample]:
                error = True
            if "missing_warning" in data[sample]:
                warning = True
            if warning and error:
                break
        return warning, error



# -*- coding: utf-8 -*-

""" MultiQC submodule to parse output from GATK varianteval """

import logging

from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class VariantEvalMixin:
    def parse_gatk_varianteval(self):
        """Find GATK varianteval logs and parse their data"""

        self.gatk_varianteval = dict()
        for f in self.find_log_files("gatk/varianteval", filehandles=True):
            parsed_data = parse_single_report(f["f"])
            if len(parsed_data) > 1:
                if f["s_name"] in self.gatk_varianteval:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="varianteval")
                self.gatk_varianteval[f["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.gatk_varianteval = self.ignore_samples(self.gatk_varianteval)

        n_reports_found = len(self.gatk_varianteval)
        if n_reports_found == 0:
            return 0

        log.info(f"Found {n_reports_found} VariantEval reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file (restructure first)
        self.write_data_file(self.gatk_varianteval, "multiqc_gatk_varianteval")

        # Get consensus TiTv references
        titv_ref = None
        for s_name in self.gatk_varianteval:
            if titv_ref is None:
                titv_ref = self.gatk_varianteval[s_name]["titv_reference"]
            elif titv_ref != self.gatk_varianteval[s_name]["titv_reference"]:
                titv_ref = "Multiple"
                break

        # General Stats Table
        varianteval_headers = dict()
        varianteval_headers["known_titv"] = {
            "title": "TiTV ratio (known)",
            "description": f"TiTV ratio from variants found in '{titv_ref}'",
            "min": 0,
            "scale": "Blues",
            "shared_key": "titv_ratio",
        }
        varianteval_headers["novel_titv"] = {
            "title": "TiTV ratio (novel)",
            "description": f"TiTV ratio from variants NOT found in '{titv_ref}'",
            "min": 0,
            "scale": "Blues",
            "shared_key": "titv_ratio",
        }
        varianteval_headers["called_titv"] = {
            "title": "TiTV ratio (called)",
            "description": f"TiTV ratio from variants found in '{titv_ref}'",
            "min": 0,
            "scale": "Blues",
            "shared_key": "titv_ratio",
        }
        varianteval_headers["filtered_titv"] = {
            "title": "TiTV ratio (filtered)",
            "description": f"TiTV ratio from variants NOT found in '{titv_ref}'",
            "min": 0,
            "scale": "Blues",
            "shared_key": "titv_ratio",
        }
        self.general_stats_addcols(self.gatk_varianteval, varianteval_headers, "GATK VariantEval")

        # Variant Counts plot
        self.add_section(
            name="Variant Counts", anchor="gatk-count-variants", plot=count_variants_barplot(self.gatk_varianteval)
        )

        # Compare Overlap Table
        self.add_section(
            name="Compare Overlap", anchor="gatk-compare-overlap", plot=comp_overlap_table(self.gatk_varianteval)
        )

        return n_reports_found


def parse_single_report(f):
    """Parse a gatk varianteval varianteval"""
    # Fixme: Separate GATKReport parsing and data subsetting. A GATKReport parser now available from the GATK MultiqcModel.

    data = dict()
    in_CompOverlap = False
    in_CountVariants = False
    in_TiTv = False
    for line in f:
        # Detect section headers
        if "#:GATKTable:CompOverlap" in line:
            in_CompOverlap = True
        elif "#:GATKTable:CountVariants" in line:
            in_CountVariants = True
        elif "#:GATKTable:TiTvVariantEvaluator" in line:
            in_TiTv = True
        else:
            # Parse contents using nested loops
            if in_CompOverlap:
                headers = line.split()
                while in_CompOverlap:
                    line = f.readline().strip("\n")
                    if len(line) < len(headers):
                        in_CompOverlap = False
                        break
                    d = dict()
                    for i, s in enumerate(line.split()):
                        d[headers[i]] = s
                    if d.get("Novelty") == "all" or d.get("Filter") == "raw":
                        if "CompFeatureInput" in d:
                            data["reference"] = d["CompFeatureInput"]
                        if "CompRod" in d:
                            data["reference"] = d["CompRod"]
                        if "compRate" in d:
                            data["comp_rate"] = float(d["compRate"])
                        if "concordantRate" in d:
                            data["concordant_rate"] = float(d["concordantRate"])
                        if "nEvalVariants" in d:
                            data["eval_variants"] = int(d["nEvalVariants"])
                        if "novelSites" in d:
                            data["novel_sites"] = int(d["novelSites"])
                    elif d.get("Novelty") == "known":
                        if "nEvalVariants" in d:
                            data["known_sites"] = int(d["nEvalVariants"])

            elif in_CountVariants:
                headers = line.split()
                while in_CountVariants:
                    line = f.readline().strip("\n")
                    if len(line) < len(headers):
                        in_CountVariants = False
                        break
                    d = dict()
                    for i, s in enumerate(line.split()):
                        d[headers[i]] = s
                    if d.get("Novelty") == "all" or d.get("Filter") == "raw":
                        if "nSNPs" in d:
                            data["snps"] = int(d["nSNPs"])
                        if "nMNPs" in d:
                            data["mnps"] = int(d["nMNPs"])
                        if "nInsertions" in d:
                            data["insertions"] = int(d["nInsertions"])
                        if "nDeletions" in d:
                            data["deletions"] = int(d["nDeletions"])
                        if "nComplex" in d:
                            data["complex"] = int(d["nComplex"])
                        if "nSymbolic" in d:
                            data["symbolic"] = int(d["nSymbolic"])
                        if "nMixed" in d:
                            data["mixed"] = int(d["nMixed"])
                        if "nNoCalls" in d:
                            data["nocalls"] = int(d["nNoCalls"])
            elif in_TiTv:
                headers = line.split()
                data["titv_reference"] = "unknown"
                while in_TiTv:
                    line = f.readline().strip("\n")
                    if len(line) < len(headers):
                        in_TiTv = False
                        break
                    d = dict()
                    for i, s in enumerate(line.split()):
                        d[headers[i]] = s
                    if d.get("Novelty") == "known":
                        if "CompFeatureInput" in d:
                            data["titv_reference"] = d["CompFeatureInput"]
                        if "CompRod" in d:
                            data["titv_reference"] = d["CompRod"]
                        if "tiTvRatio" in d:
                            data["known_titv"] = float(d["tiTvRatio"])
                    elif d.get("Novelty") == "novel":
                        if "tiTvRatio" in d:
                            data["novel_titv"] = float(d["tiTvRatio"])
                    elif d.get("Filter") == "called":
                        if "CompFeatureInput" in d:
                            data["titv_reference"] = d["CompFeatureInput"]
                        if "CompRod" in d:
                            data["titv_reference"] = d["CompRod"]
                        if "tiTvRatio" in d:
                            data["called_titv"] = float(d["tiTvRatio"])
                    elif d.get("Filter") == "filtered":
                        if "CompFeatureInput" in d:
                            data["titv_reference"] = d["CompFeatureInput"]
                        if "CompRod" in d:
                            data["titv_reference"] = d["CompRod"]
                        if "tiTvRatio" in d:
                            data["filtered_titv"] = float(d["tiTvRatio"])

    return data


def count_variants_barplot(data):
    """Return HTML for the Variant Counts barplot"""
    keys = {
        "snps": {"name": "SNPs"},
        "mnps": {"name": "MNPs"},
        "insertions": {"name": "Insertions"},
        "deletions": {"name": "Deletions"},
        "complex": {"name": "Complex"},
        "symbolic": {"name": "Symbolic"},
        "mixed": {"name": "Mixed"},
        "nocalls": {"name": "No-calls"},
    }

    plot_conf = {
        "id": "gatk_varianteval_variant_plot",
        "title": "GATK VariantEval: Variant Counts",
        "ylab": "# Variants",
        "cpswitch_counts_label": "Number of Variants",
    }
    return bargraph.plot(data, keys, plot_conf)


def comp_overlap_table(data):
    """Build a table from the comp overlaps output."""
    headers = {
        "comp_rate": {
            "title": "Compare rate",
            "description": "Ratio of known variants found in the reference set.",
            "namespace": "GATK",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "format": "{:,.2f}",
            "scale": "Blues",
        },
        "concordant_rate": {
            "title": "Concordant rate",
            "description": "Ratio of variants matching alleles in the reference set.",
            "namespace": "GATK",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "format": "{:,.2f}",
            "scale": "Blues",
        },
        "eval_variants": {
            "title": "M Evaluated variants",
            "description": "Number of called variants (millions)",
            "namespace": "GATK",
            "min": 0,
            "modify": lambda x: float(x) / 1000000.0,
        },
        "known_sites": {
            "title": "M Known sites",
            "description": "Number of known variants (millions)",
            "namespace": "GATK",
            "min": 0,
            "modify": lambda x: float(x) / 1000000.0,
        },
        "novel_sites": {
            "title": "M Novel sites",
            "description": "Number of novel variants (millions)",
            "namespace": "GATK",
            "min": 0,
            "modify": lambda x: float(x) / 1000000.0,
        },
    }
    table_html = table.plot(data, headers, {"id": "gatk_compare_overlap", "table_title": "GATK - Compare Overlap"})
    return table_html



#!/usr/bin/env python

""" MultiQC functions to plot a beeswarm group """

import logging
import random
from collections import defaultdict

from multiqc.plots import table_object
from multiqc.utils import config, report, util_functions

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"


# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(dt: table_object.DataTable):
    """
    Make a plot - template custom, or interactive or flat
    """
    categories = []
    s_names = []
    data = []
    dt.raw_vals = defaultdict(lambda: dict())
    for idx, hs in enumerate(dt.headers):
        for k, header in hs.items():
            bcol = f"rgb({header.get('colour', '204,204,204')})"

            categories.append(
                {
                    "namespace": header["namespace"],
                    "title": header["title"],
                    "description": header["description"],
                    "max": header["dmax"],
                    "min": header["dmin"],
                    "suffix": header.get("suffix", ""),
                    "decimalPlaces": header.get("decimalPlaces", "2"),
                    "bordercol": bcol,
                }
            )

            # Add the data
            thisdata = []
            these_snames = []
            for s_name, samp in dt.data[idx].items():
                if k in samp:
                    val = samp[k]
                    dt.raw_vals[s_name][k] = val

                    if "modify" in header and callable(header["modify"]):
                        val = header["modify"](val)

                    thisdata.append(val)
                    these_snames.append(s_name)

            data.append(thisdata)
            s_names.append(these_snames)

    if len(s_names) == 0:
        logger.warning("Tried to make beeswarm plot, but had no data")
        return '<p class="text-danger">Error - was not able to plot data.</p>'

    bs_id = dt.pconfig.get("id", f"table_{''.join(random.sample(letters, 4))}")

    # Sanitise plot ID and check for duplicates
    bs_id = report.save_htmlid(bs_id)

    # Plot HTML
    html = """<div class="hc-plot-wrapper"{height}>
        <div id="{bid}" class="hc-plot not_rendered hc-beeswarm-plot"><small>loading..</small></div>
    </div>""".format(
        bid=bs_id,
        height=f' style="height:{dt.pconfig["height"]}px"' if "height" in dt.pconfig else "",
    )

    report.num_hc_plots += 1

    report.plot_data[bs_id] = {"plot_type": "beeswarm", "samples": s_names, "datasets": data, "categories": categories}

    # Save the raw values to a file if requested
    if dt.pconfig.get("save_file") is True:
        fn = dt.pconfig.get("raw_data_fn", f"multiqc_{bs_id}")
        util_functions.write_data_file(dt.raw_vals, fn)
        report.saved_raw_data[fn] = dt.raw_vals

    return html



""" MultiQC modules base class, contains helper functions """
from typing import List, Union, Optional

import fnmatch
import io
import itertools
import logging
import mimetypes
import os
import re
import textwrap
from collections import defaultdict

import markdown

from multiqc.utils import config, report, software_versions, util_functions

logger = logging.getLogger(__name__)


class ModuleNoSamplesFound(Exception):
    """Module checked all input files but couldn't find any data to use"""


class BaseMultiqcModule(object):
    def __init__(
        self,
        name="base",
        anchor="base",
        target=None,
        href=None,
        info=None,
        comment=None,
        extra=None,
        autoformat=True,
        autoformat_type="markdown",
        doi=None,
    ):
        # Custom options from user config that can overwrite base module values
        mod_cust_config = getattr(self, "mod_cust_config", {})
        self.name = mod_cust_config.get("name", name)
        self.anchor = mod_cust_config.get("anchor", anchor)
        target = mod_cust_config.get("target", target)
        self.href = mod_cust_config.get("href", href)
        self.info = mod_cust_config.get("info", info)
        self.comment = mod_cust_config.get("comment", comment)
        self.extra = mod_cust_config.get("extra", extra)
        self.doi = mod_cust_config.get("doi", (doi or []))

        # List of software version(s) for module. Don't append directly, use add_software_version()
        self.versions = defaultdict(list)

        # Specific module level config to overwrite (e.g. config.bcftools, config.fastqc)
        config.update({anchor: mod_cust_config.get("custom_config", {})})

        # Sanitise anchor ID and check for duplicates
        self.anchor = report.save_htmlid(self.anchor)

        # See if we have a user comment in the config
        if self.anchor in config.section_comments:
            self.comment = config.section_comments[self.anchor]

        if self.info is None:
            self.info = ""
        # Always finish with a ".", as we may add a DOI after the intro.
        if len(self.info) > 0 and self.info[-1] != ".":
            self.info += "."
        if self.extra is None:
            self.extra = ""
        self.doi_link = ""
        if isinstance(self.doi, str):
            self.doi = [self.doi]
        self.doi = [i for i in self.doi if i != ""]
        if len(self.doi) > 0:
            doi_links = []
            for doi in self.doi:
                # Build the HTML link for the DOI
                doi_links.append(
                    f' <a class="module-doi" data-doi="{doi}" data-toggle="popover" href="https://doi.org/{doi}" target="_blank">{doi}</a>'
                )
            self.doi_link = '<em class="text-muted small" style="margin-left: 1rem;">DOI: {}.</em>'.format(
                "; ".join(doi_links)
            )

        if target is None:
            target = self.name
        if self.href is not None:
            self.mname = f'<a href="{self.href}" target="_blank">{target}</a>'
        else:
            self.mname = target
        if self.href or self.info or self.extra or self.doi_link:
            self.intro = f"<p>{self.mname} {self.info}{self.doi_link}</p>{self.extra}"

        # Format the markdown strings
        if autoformat:
            if self.comment is not None:
                self.comment = textwrap.dedent(self.comment)
                if autoformat_type == "markdown":
                    self.comment = markdown.markdown(self.comment)

        self.sections = list()

    def find_log_files(self, sp_key, filecontents=True, filehandles=False):
        """
        Return matches log files of interest.
        :param sp_key: Search pattern key specified in config
        :param filehandles: Set to true to return a file handle instead of slurped file contents
        :return: Yields a dict with filename (fn), root directory (root), cleaned sample name
                 generated from the filename (s_name) and either the file contents or file handle
                 for the current matched file (f).
                 As yield is used, the results can be iterated over without loading all files at once
        """

        # Pick up path filters if specified.
        # Allows modules to be called multiple times with different sets of files
        path_filters = getattr(self, "mod_cust_config", {}).get("path_filters")
        path_filters_exclude = getattr(self, "mod_cust_config", {}).get("path_filters_exclude")
        if isinstance(path_filters, str):
            path_filters = [path_filters]
        if isinstance(path_filters_exclude, str):
            path_filters_exclude = [path_filters_exclude]

        # Old, depreciated syntax support. Likely to be removed in a future version.
        if isinstance(sp_key, dict):
            report.files[self.name] = list()
            for sf in report.searchfiles:
                if report.search_file(sp_key, {"fn": sf[0], "root": sf[1]}, module_key=None):
                    report.files[self.name].append({"fn": sf[0], "root": sf[1]})
            sp_key = self.name
            logwarn = f"Depreciation Warning: {self.name} - Please use new style for find_log_files()"
            if len(report.files[self.name]) > 0:
                logger.warning(logwarn)
            else:
                logger.debug(logwarn)
        elif not isinstance(sp_key, str):
            logger.warning("Did not understand find_log_files() search key")
            return

        for f in report.files[sp_key]:
            # Make a note of the filename so that we can report it if something crashes
            report.last_found_file = os.path.join(f["root"], f["fn"])

            # Filter out files based on exclusion patterns
            if path_filters_exclude and len(path_filters_exclude) > 0:
                # Try both the given path and also the path prefixed with the analysis dirs
                exlusion_hits = itertools.chain(
                    (fnmatch.fnmatch(report.last_found_file, pfe) for pfe in path_filters_exclude),
                    *(
                        (
                            fnmatch.fnmatch(report.last_found_file, os.path.join(analysis_dir, pfe))
                            for pfe in path_filters_exclude
                        )
                        for analysis_dir in config.analysis_dir
                    ),
                )
                if any(exlusion_hits):
                    logger.debug(
                        f"{sp_key} - Skipping '{report.last_found_file}' as it matched the path_filters_exclude for '{self.name}'"
                    )
                    continue

            # Filter out files based on inclusion patterns
            if path_filters and len(path_filters) > 0:
                # Try both the given path and also the path prefixed with the analyis dirs
                inclusion_hits = itertools.chain(
                    (fnmatch.fnmatch(report.last_found_file, pf) for pf in path_filters),
                    *(
                        (fnmatch.fnmatch(report.last_found_file, os.path.join(analysis_dir, pf)) for pf in path_filters)
                        for analysis_dir in config.analysis_dir
                    ),
                )
                if not any(inclusion_hits):
                    logger.debug(
                        f"{sp_key} - Skipping '{report.last_found_file}' as it didn't match the path_filters for '{self.name}'"
                    )
                    continue
                else:
                    logger.debug(
                        f"{sp_key} - Selecting '{report.last_found_file}' as it matched the path_filters for '{self.name}'"
                    )

            # Make a sample name from the filename
            f["sp_key"] = sp_key
            f["s_name"] = self.clean_s_name(f["fn"], f)
            if filehandles or filecontents:
                try:
                    # Custom content module can now handle image files
                    (ftype, encoding) = mimetypes.guess_type(os.path.join(f["root"], f["fn"]))
                    if ftype is not None and ftype.startswith("image"):
                        with io.open(os.path.join(f["root"], f["fn"]), "rb") as fh:
                            # always return file handles
                            f["f"] = fh
                            yield f
                    else:
                        # Everything else - should be all text files
                        with io.open(os.path.join(f["root"], f["fn"]), "r", encoding="utf-8") as fh:
                            if filehandles:
                                f["f"] = fh
                                yield f
                            elif filecontents:
                                try:
                                    f["f"] = fh.read()
                                except UnicodeDecodeError as e:
                                    logger.debug(
                                        f"Couldn't read file as utf-8: {f['fn']}, will attempt to skip non-unicode characters\n{e}"
                                    )
                                    try:
                                        with io.open(
                                            os.path.join(f["root"], f["fn"]), "r", encoding="utf-8", errors="ignore"
                                        ) as fh_ignoring:
                                            f["f"] = fh_ignoring.read()
                                    except Exception as e:
                                        logger.debug(f"Still couldn't read file: {f['fn']}\n{e}")
                                        f["f"] = None
                                    finally:
                                        fh.close()
                                yield f
                except (IOError, OSError, ValueError, UnicodeDecodeError) as e:
                    logger.debug(f"Couldn't open filehandle when returning file: {f['fn']}\n{e}")
                    f["f"] = None
            else:
                yield f

    def add_section(
        self,
        name=None,
        anchor=None,
        description="",
        comment="",
        helptext="",
        plot="",
        content="",
        autoformat=True,
        autoformat_type="markdown",
    ):
        """Add a section to the module report output"""

        # Default anchor
        if anchor is None:
            if name is not None:
                nid = name.lower().strip().replace(" ", "-")
                anchor = f"{self.anchor}-{nid}"
            else:
                sl = len(self.sections) + 1
                anchor = f"{self.anchor}-section-{sl}"

        # Append custom module anchor to the section if set
        mod_cust_config = getattr(self, "mod_cust_config", {})
        if "anchor" in mod_cust_config:
            anchor = f"{mod_cust_config['anchor']}_{anchor}"

        # Sanitise anchor ID and check for duplicates
        anchor = report.save_htmlid(anchor)

        # Skip if user has a config to remove this module section
        if anchor in config.remove_sections:
            logger.debug(f"Skipping section '{anchor}' because specified in user config")
            return

        # See if we have a user comment in the config
        if anchor in config.section_comments:
            comment = config.section_comments[anchor]

        # Format the content
        if autoformat:
            if len(description) > 0:
                description = textwrap.dedent(description)
                if autoformat_type == "markdown":
                    description = markdown.markdown(description)
            if len(comment) > 0:
                comment = textwrap.dedent(comment)
                if autoformat_type == "markdown":
                    comment = markdown.markdown(comment)
            if len(helptext) > 0:
                helptext = textwrap.dedent(helptext)
                if autoformat_type == "markdown":
                    helptext = markdown.markdown(helptext)

        # Strip excess whitespace
        description = description.strip()
        comment = comment.strip()
        helptext = helptext.strip()

        self.sections.append(
            {
                "name": name,
                "anchor": anchor,
                "description": description,
                "comment": comment,
                "helptext": helptext,
                "plot": plot,
                "content": content,
                "print_section": any(
                    [n is not None and len(n) > 0 for n in [description, comment, helptext, plot, content]]
                ),
            }
        )

    @staticmethod
    def _clean_fastq_pair(r1: str, r2: str) -> Optional[str]:
        """
        Try trimming r1 and r2 as paired FASTQ file names.
        """
        # Try trimming the conventional illumina suffix with a tail 001 ending. Refs:
        # https://support.illumina.com/help/BaseSpace_Sequence_Hub_OLH_009008_2/Source/Informatics/BS/NamingConvention_FASTQ-files-swBS.htm
        # https://support.10xgenomics.com/spatial-gene-expression/software/pipelines/latest/using/fastq-input#:~:text=10x%20pipelines%20need%20files%20named,individual%20who%20demultiplexed%20your%20flowcell.
        cleaned_r1 = re.sub(r"_R1_\d{3}$", "", r1)
        cleaned_r2 = re.sub(r"_R2_\d{3}$", "", r2)
        if cleaned_r1 == cleaned_r2:  # trimmed successfully
            return cleaned_r1

        # Try removing _R1 and _R2 from the middle.
        cleaned_r1 = re.sub(r"_R1_", "_", r1)
        cleaned_r2 = re.sub(r"_R2_", "_", r2)
        if cleaned_r1 == cleaned_r2:  # trimmed successfully
            return cleaned_r1

        # Try trimming other variations from the end (-R1, _r1, _1, .1, etc).
        cleaned_r1 = re.sub(r"([_.-][rR]?1)?$", "", r1)
        cleaned_r2 = re.sub(r"([_.-][rR]?2)?$", "", r2)
        if cleaned_r1 == cleaned_r2:  # trimmed successfully
            return cleaned_r1

        return None

    def clean_s_name(self, s_name: Union[str, List[str]], f=None, root=None, filename=None, seach_pattern_key=None):
        """
        Helper function to take a long file name(s) and strip back to one clean sample name. Somewhat arbitrary.
        :param s_name: The sample name(s) to clean.
        :param root: The directory path that this file is within
        :config.prepend_dirs: boolean, whether to prepend dir name to s_name
        :return: The cleaned sample name, ready to be used
        """
        if isinstance(s_name, list):
            if len(s_name) == 0:
                raise ValueError("Empty list of sample names passed to clean_s_name()")

            # Extract a sample name from a list of file names (for example, FASTQ pairs).
            # Each name is cleaned separately first:
            clean_names = [
                self.clean_s_name(sn, f=f, root=root, filename=filename, seach_pattern_key=seach_pattern_key)
                for sn in s_name
            ]
            if len(set(clean_names)) == 1:
                # All the same, returning the first one.
                return clean_names[0]

            if len(clean_names) == 2:
                # Checking if it's a FASTQ pair.
                fastq_s_name = self._clean_fastq_pair(*clean_names)
                if fastq_s_name is not None:
                    return fastq_s_name

            # Couldn't clean as FASTQ. Just concatenating the clean names.
            return "_".join(clean_names)

        s_name_original = s_name

        # Backwards compatability - if f is a string, it's probably the root (this used to be the second argument)
        if isinstance(f, str):
            root = f
            f = None

        # Set string variables from f if it was a dict from find_log_files()
        if isinstance(f, dict):
            if "root" in f and root is None:
                root = f["root"]
            if "fn" in f and filename is None:
                filename = f["fn"]
            if "sp_key" in f and seach_pattern_key is None:
                seach_pattern_key = f["sp_key"]

        # For modules setting s_name from file contents, set s_name back to the filename
        # (if wanted in the config)
        if filename is not None and (
            config.use_filename_as_sample_name is True
            or (
                isinstance(config.use_filename_as_sample_name, list)
                and seach_pattern_key is not None
                and seach_pattern_key in config.use_filename_as_sample_name
            )
        ):
            s_name = filename

        # Set root to empty string if not known
        if root is None:
            root = ""

        # if s_name comes from file contents, it may have a file path
        # For consistency with other modules, we keep just the basename
        s_name = os.path.basename(s_name)

        # Prepend sample name with directory
        if config.prepend_dirs:
            sep = config.prepend_dirs_sep
            root = root.lstrip(f".{os.sep}")
            dirs = [d.strip() for d in root.split(os.sep) if d.strip() != ""]
            if config.prepend_dirs_depth != 0:
                d_idx = config.prepend_dirs_depth * -1
                if config.prepend_dirs_depth > 0:
                    dirs = dirs[d_idx:]
                else:
                    dirs = dirs[:d_idx]
            if len(dirs) > 0:
                s_name = f"{sep.join(dirs)}{sep}{s_name}"

        if config.fn_clean_sample_names:
            # Split then take first section to remove everything after these matches
            for ext in config.fn_clean_exts:
                # Check if this config is limited to a module
                if "module" in ext:
                    if isinstance(ext["module"], str):
                        ext["module"] = [ext["module"]]
                    if not any([m == self.anchor for m in ext["module"]]):
                        continue

                # Go through different filter types
                if isinstance(ext, str):
                    ext = {"type": "truncate", "pattern": ext}
                if ext.get("type") == "truncate":
                    s_name = s_name.split(ext["pattern"], 1)[0]
                elif ext.get("type") in ("remove", "replace"):
                    if ext["type"] == "replace":
                        logger.warning(
                            "use 'config.fn_clean_sample_names.remove' instead "
                            "of 'config.fn_clean_sample_names.replace' [deprecated]"
                        )
                    s_name = s_name.replace(ext["pattern"], "")
                elif ext.get("type") == "regex":
                    s_name = re.sub(ext["pattern"], "", s_name)
                elif ext.get("type") == "regex_keep":
                    match = re.search(ext["pattern"], s_name)
                    s_name = match.group() if match else s_name
                elif ext.get("type") is None:
                    logger.error(f'config.fn_clean_exts config was missing "type" key: {ext}')
                else:
                    logger.error(f"Unrecognised config.fn_clean_exts type: {ext.get('type')}")
            # Trim off characters at the end of names
            for chrs in config.fn_clean_trim:
                if s_name.endswith(chrs):
                    s_name = s_name[: -len(chrs)]
                if s_name.startswith(chrs):
                    s_name = s_name[len(chrs) :]

        # Remove trailing whitespace
        s_name = s_name.strip()

        # If we cleaned back to an empty string, just use the original value
        if s_name == "":
            s_name = s_name_original

        # Do any hard replacements that are set with --replace-names
        if config.sample_names_replace:
            for s_name_search, s_name_replace in config.sample_names_replace.items():
                try:
                    # Skip if we're looking for exact matches only
                    if config.sample_names_replace_exact:
                        # Simple strings
                        if not config.sample_names_replace_regex and s_name != s_name_search:
                            continue
                        # regexes
                        if config.sample_names_replace_regex and not re.fullmatch(s_name_search, s_name):
                            continue
                    # Replace - regex
                    if config.sample_names_replace_regex:
                        s_name = re.sub(s_name_search, s_name_replace, s_name)
                    # Replace - simple string
                    else:
                        # Complete name swap
                        if config.sample_names_replace_complete:
                            if s_name_search in s_name:
                                s_name = s_name_replace
                        # Partial substring replace
                        else:
                            s_name = s_name.replace(s_name_search, s_name_replace)
                except re.error as e:
                    logger.error(f"Error with sample name replacement regex: {e}")

        return s_name

    def ignore_samples(self, data):
        """Strip out samples which match `sample_names_ignore`"""
        try:
            if isinstance(data, dict):
                newdata = dict()
            else:
                return data
            for s_name, v in data.items():
                if not self.is_ignore_sample(s_name):
                    newdata[s_name] = v
            return newdata
        except (TypeError, AttributeError):
            return data

    def is_ignore_sample(self, s_name):
        """Should a sample name be ignored?"""
        glob_match = any(fnmatch.fnmatch(s_name, sn) for sn in config.sample_names_ignore)
        re_match = any(re.match(sn, s_name) for sn in config.sample_names_ignore_re)
        return glob_match or re_match

    def general_stats_addcols(self, data, headers=None, namespace=None):
        """Helper function to add to the General Statistics variable.
        Adds to report.general_stats and does not return anything. Fills
        in required config variables if not supplied.
        :param data: A dict with the data. First key should be sample name,
                     then the data key, then the data.
        :param headers: Dict with information for the headers,
                        such as colour scales, min and max values etc.
                        See docs/writing_python.md for more information.
        :param namespace: Append to the module name in the table column description.
                          Can be e.g. a submodule name.
        :return: None
        """
        if headers is None:
            headers = {}
        # Deepish copy of headers so that we can modify it in place
        headers = {k: v.copy() for k, v in headers.items()}

        # Guess the column headers from the data if not supplied
        if headers is None or len(headers) == 0:
            hs = set()
            for d in data.values():
                hs.update(d.keys())
            hs = list(hs)
            hs.sort()
            headers = dict()
            for k in hs:
                headers[k] = dict()

        # Add the module name to the description if not already done
        keys = headers.keys()
        for k in keys:
            # Prepend the namespace displayed in the table with the module name
            namespace = headers[k].get("namespace", namespace)
            headers[k]["namespace"] = self.name
            if namespace:
                headers[k]["namespace"] = self.name + ": " + namespace
            if "description" not in headers[k]:
                headers[k]["description"] = headers[k].get("title", k)

        # Append to report.general_stats for later assembly into table
        report.general_stats_data.append(data)
        report.general_stats_headers.append(headers)

    def add_data_source(self, f=None, s_name=None, source=None, module=None, section=None):
        try:
            if module is None:
                module = self.name
            if section is None:
                section = "all_sections"
            if s_name is None:
                s_name = f["s_name"]
            if source is None:
                source = os.path.abspath(os.path.join(f["root"], f["fn"]))
            report.data_sources[module][section][s_name] = source
        except AttributeError:
            logger.warning(f"Tried to add data source for {self.name}, but was missing fields data")

    def add_software_version(self, version: str = None, sample: str = None, software_name: str = None):
        """Save software versions for module."""
        # Don't add if version is None. This allows every module to call this function
        # even those without a version to add. This is useful to check that all modules
        # are calling this function.
        if version is None:
            return

        # Don't add if version detection is disabled
        if config.disable_version_detection:
            return

        # Don't add if sample is ignored
        if sample is not None and self.is_ignore_sample(sample):
            return

        # Use module name as software name if not specified
        if software_name is None:
            software_name = self.name

        # Check if version string is PEP 440 compliant to enable version normalization and proper ordering.
        # Otherwise, use raw string is used for version.
        # - https://peps.python.org/pep-0440/
        version = software_versions.parse_version(version)

        if version in self.versions[software_name]:
            return

        self.versions[software_name].append(version)

        # Sort version in order newest --> oldest
        self.versions[software_name] = software_versions.sort_versions(self.versions[software_name])

        # Update version list for report section.
        group_name = self.name
        report.software_versions[group_name][software_name] = self.versions[software_name]

    def write_data_file(self, data, fn, sort_cols=False, data_format=None):
        """Saves raw data to a dictionary for downstream use, then redirects
        to report.write_data_file() to create the file in the report directory"""

        # Append custom module anchor if set
        mod_cust_config = getattr(self, "mod_cust_config", {})
        if "anchor" in mod_cust_config:
            fn = f"{fn}_{mod_cust_config['anchor']}"

        # Generate a unique filename if the file already exists (running module multiple times)
        i = 1
        base_fn = fn
        while fn in report.saved_raw_data:
            fn = f"{base_fn}_{i}"
            i += 1

        # Save the file
        report.saved_raw_data[fn] = data
        util_functions.write_data_file(data, fn, sort_cols, data_format)

    ##################################################
    #### DEPRECATED FORWARDERS
    def plot_bargraph(self, data, cats=None, pconfig=None):
        """Depreciated function. Forwards to new location."""
        from multiqc.plots import bargraph

        if pconfig is None:
            pconfig = {}
        return bargraph.plot(data, cats, pconfig)

    def plot_xy_data(self, data, pconfig=None):
        """Depreciated function. Forwards to new location."""
        from multiqc.plots import linegraph

        if pconfig is None:
            pconfig = {}
        return linegraph.plot(data, pconfig)
